/// \file standalone.c -- VS1063 Stand-Alone MP3 Player Source Code
/// Copyright (C) 2011-2017 VLSI Solution Oy Tampere Finland

/*
  UART_OUTPUT configures the encoder to send the data to UART instead
  of saving to and creating a file.
 */
//#define UART_OUTPUT 22050U //Only for stereo line-in.. do not make a file!!

/*
  Note:

  DREQ is changed in normal firmware both in SDI interrupts when the
  stream buffer becomes too full, and whenever the stream buffer becomes
  empty enough.

  So, we can use DREQ ourselves as we wish and access to it does not need to
  be protected by interrupt Disable()/Enable(). However, the normal decoding
  will still keep raising DREQ when it sees that there is space for new data.

- VS1063:
  + I2S
  + higher reference
- IN VS1063:
  + SCLK = GPIDATA8
  + XCS  = GPIDATA9
  + SI   = GPIDATA10
  + XDCS = GPIDATA11


- Exiting recoding now from prev/next buttons. Rec pause/continue in the middle
  button. We use a timer to get 32Hz trigger for key scan.
- Now ends recording nicely and creates 100% valid files.
- If format is WAV, fix the WAV header.

 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <h1063/encoders.h>
#include "standalone.h"
int csCancel = 0;

#ifndef ENABLE_I2S /*I2S uses GPIO4..6, so they cannot be used for these. */
//#define LINE_IN_ENABLE (1<<6) /* GPIO6='1' selects stereo line, otherwise mic  +18words*/
//#define RECORDING_INDICATOR (1<<4) /*+7words*/
//#define PREPARE_RECORDING_INDICATOR (1<<5)
#endif

/*
  USE_READ_MULTIPLE_BLOCK_ONE is a workaround for
  1. some cards that have issues with implementing READ_SINGLE_BLOCK
     in SPI mode correctly.
  2. cards that have issues implementing READ_MULTIPLE_BLOCK in SPI mode
     correctly.

  Use READ_MULTIPLE_BLOCK, but only read one block at a time.

  See also: asm-readdisk.s
 */
#define USE_READ_MULTIPLE_BLOCK_ONE /* Workaround for some cards that fail with single read block. Is the same size, so unless you use old MMC cards that do not properly support multiple block read, you can keep this defined. If you comment this out, also change #if 0 in asm-readdisk.s to #if 1 */


#ifdef UART_BUFFERED
/*Recorder needs currently more code space... Disable some features.*/
#define NO_LOOP_MODE
#define NO_RANDOM
#define NO_DREQ_LED
  #ifdef SHUFFLE_PLAY
    #undef SHUFFLE_PLAY
  #endif
  #ifdef SKIP_ID3V2
    #undef SKIP_ID3V2
  #endif
#endif /*UART_BUFFERED*/

/*
  If any of the key control alternatives, enable other key-related stuff.
 */
#if defined(SCI_KEYS) || defined(COMPAT_KEYS) || defined(DIRECT_KEYS)
  #define LONG_PRESS_TIME 24
  #define SW1 0xffffU
  #define SW2 0x8000U
  #define SW3 0x7fffU

  #define HAS_KEYS
#endif

//u_int16 sectorBuffer[512/2] = {0};

//#define USE_DEBUG

/* See standalone.h for feature #defines. */

#if defined(SAVE_VOLUME) || defined(SAVE_POSITION)
  #define SAVE_STUFF
#endif

extern volatile s_int16 *srcWp, *srcRp;

extern __y u_int16 g_dcthi[2048];
auto void play_frame(void);
//void putch(register __a0 int ch);
void putword(register __a0 int ch);

#define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))

s_int16 hcShift; /*SDHC*/

#ifndef GPIO_IDATA
  #define GPIO_IDATA GPIO_DATA
#endif

#define SPI_xCS   1 /* GPIO */
#define SPI_CLK   8
#define SPI_MISO  4
#define SPI_MISO_SHIFT -3
#define SPI_xCS2  2
#define SPI_MOSI  1 /* DREQ */

#ifdef UART_BUFFERED
#ifndef NO_FILEMODE
#define USE_FILEMODE
#endif
#endif

#ifdef USE_FILEMODE
static int fileMode =
#ifdef START_IN_FILEMODE
  1
#else
  0
#endif
  ;
#endif/*USE_FILEMODE*/

typedef enum {
    spi_pause=0,
    spi_file=1,
    spi_seek=2,
    spi_waitdata=3,
    spi_data=4,
    spi_zeros=5,
    spi_record=6,
    spi_waitima=7
} SPISTATE;

SPISTATE spiState, oldSpiState;
u_int16 spiCnt;


struct SPIX { /* do not touch! ASM uses the structure */
    s_int32 size;
    u_int32 sector;
    u_int16 fragSize; /* TODO: make this 32-bit and multiply fragment list clusters by sectorsPerCluster. */
    __mem_y struct FRAGMENT *fragment;
} spix;


u_int16 playbackVol; /*temporary store for volume during encoding mode*/
struct REC {
    u_int16 record;
    u_int16 rd;
    u_int16 wr;
    s_int16 fill;
} record;
//free_y is 0xfa14..0xffff
#define YBUF ((__y s_int16 *)0xe000)
#define YBUFSZ 0x1800 //n*256 0xe000..0xf7ff


extern __y u_int16 hwSampleRate;

/* Can be done for all formats, because IdleHook is run as the main thread
   and not in interrupts. Whenever IdleHook is called, there either is
   not enough stream data anyway, or the decoder is outputting samples
   and there is not enough space in the audio buffer (thus not reading
   stream). */
u_int16 __x * __x stream_save;

#ifdef HAS_KEYS
__mem_y struct {
    u_int16 Data; /* must be first entry! Used by c.s */
    u_int16 Cnt;
    u_int16 Strobe;
    u_int16 Timer; /* Data,Cnt,Strobe,Timer must be in this order (buts.s) */
    u_int16 Key;
    s_int16 KeyTime;
} ui;
#endif/*HAS_KEYS*/



void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
#ifdef SAVE_VOLUME
int volChanged = 0;
#endif

#ifdef SAVE_STUFF
  #define SPI_WREN  0x06
  #define SPI_WRDI  0x04
  #define SPI_RDSR  0x05
  #define SPI_WRSR  0x01
  #define SPI_READ  0x03
  #define SPI_WRITE 0x02
  
  register __a1 short SpiDelay(void);
  auto u_int16 SpiSendReceiveSpi(register __a0 u_int16 datTopAligned, register __a1 s_int16 bits);
  
  #if 1
  void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
  #else
  void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value) {
      USEX(GPIO_ODATA) &= ~(SPI_xCS2|SPI_xCS);
      SpiSendReceiveSpi(SPI_WREN<<8, 8);
      USEX(GPIO_ODATA) |= SPI_xCS;
      USEX(GPIO_ODATA) &= ~SPI_xCS;
      SpiSendReceiveSpi(SPI_WRITE<<8, 8);
      SpiSendReceiveSpi(addr/*volume=26*/, 16);
      SpiSendReceiveSpi(value, 16);
      USEX(GPIO_ODATA) |= SPI_xCS|SPI_xCS2;//SPI_CLK;
  }
  #endif
  //SaveWord(26/*volume=26*/, USEX(SCI_VOL));
#endif

#ifdef SHUFFLE_PLAY
u_int16 shuffle_offset = 0;
#endif

/*
  Called before HALT
 */
void UserHook(void) {
    register __c1 SPISTATE cachedSpiState = spiState;
 
    if (record.record == 0x2468) { /* In record mode */
      //putch('0'+spiState);
      USEX(SCI_AICTRL3) |= CTRL3_RECORD_ON; /* Force it on */
      USEX(SCI_AICTRL0) = 0; //to not have the top bit set!
      USEX(INT_ENABLE) |= (1<<INT_EN_RX)
#ifdef HAS_KEYS
	| (1<<INT_EN_TIM0)
#endif/*HAS_KEYS*/
	;
    }

#ifdef HAS_KEYS    
    /* -- KEY USER INTERFACE -- */
    if (ui.Strobe) {
	ui.Data = ui.Strobe = 0;
#ifdef SCI_KEYS /*If both keys and SCI control.*/
	/* SCI-controlled version, but with additional key support.
	   GPIO4 for Pause/Play(short press) and Toggle Shuffle(long press), 
	   GPIO5 for Next(short) and Vol+(long), 
	   GPIO6 for Previous(short) and Vol-(long). */
#if 1
	ui.Data = USEX(GPIO_IDATA) & (7<<4);/*low is idle, high when pressed*/
#else
	ui.Data = ~USEX(GPIO_IDATA) & (7<<4); /*high idle, low when pressed*/
#endif
	if (ui.Data == (1<<6)) {
	    ui.Data = SW2;/*prev/vol-*/
	} else if (ui.Data == (1<<5)) {
	    ui.Data = SW1;/*next/vol+*/
	} else if (ui.Data == (1<<4)) {
	    ui.Data = SW3;/*pause/play/loudness*/
	}
#endif/*SCI_KEYS*/
#ifdef COMPAT_KEYS
	/*
	  Read the 3 buttons on the VS10xx prototyping board.
	*/
	{
	  register u_int16 t = USEX(GPIO_IDATA);
	  if (!(t & (1<<10)/*SI*/)) {
	    ui.Data = SW2;/*prev/vol-*/
	  } else if (!(t & (1<<11)/*XDCS*/)) {
	    ui.Data = SW1;/*next/vol+*/
	  } else {
	    t = USEX(GPIO_ODATA);/*save state*/
	    USEX(GPIO_ODATA) = t & ~SPI_xCS;
	    /* Should have a short wait here, but seems to work this way.*/
	    USEX(GPIO_IDATA);//at least a little extra wait
	    #if 0
	    if (!(USEX(GPIO_IDATA) & (1<<11)/*XDCS*/)) {
		ui.Data = SW3;/*play/pause/ etc*/
	    } else if (!(USEX(GPIO_IDATA) & (1<<9)/*XCS*/)) {
		ui.Data = SW3;
	    }
	    #else
	    if ((USEX(GPIO_IDATA) & ((1<<11)|(1<<9))) != ((1<<11)|(1<<9))) {
	      ui.Data = SW3; /*play/pause/ etc*/
	    }
	    #endif
	    USEX(GPIO_ODATA) = t; /*restore state*/
	  }
	}
#endif/*COMPAT_KEYS*/
#ifdef DIRECT_KEYS
	/*
	  Read keys directly connected to SCI pins.
	  Idle high, button pulls low.
	  + SCLK = GPIDATA8 (unused)
	  + XCS  = GPIDATA9
	  + SI   = GPIDATA10
	  + XDCS = GPIDATA11
	*/
	if (!(USEX(GPIO_IDATA) & (1<<10)/*SI*/)) {
	    ui.Data = SW2;/*next*/
	} else if (!(USEX(GPIO_IDATA) & (1<<11)/*XDCS*/)) {
	    ui.Data = SW1;/*prev*/
	} else if (!(USEX(GPIO_IDATA) & (1<<9)/*XCS*/)) {
	    ui.Data = SW3;/*pause/play*/
	}
	//else if (!(USEX(GPIO_IDATA) & (1<<8)/*SCLK*/)) {
	//ui.Data = SW1;/*prev*/
	//}
#endif/*DIRECT_KEYS*/

	// -- KEY FUNCTIONALITY --
	if (ui.Data == ui.Key) { // New Keystate is same as previous Keystate
	    if (ui.KeyTime < 12) {
		ui.KeyTime++;
	    } else if (ui.Key && ui.KeyTime > 0) {
	      /* keystate has been consistent for 12 ticks
		 - it's a long keypress */
	      /* long presses */
	        if (ui.Key == SW1) {
		    #ifdef UART_BUFFERED
		    goto volup;
		    #else
		    if (USEX(SCI_VOL))
		      USEX(SCI_VOL) -= 0x0101;
		    goto setvol;
		    #endif
		} else if (ui.Key == SW2) {
                    #ifdef UART_BUFFERED
		    goto voldown;
		    #else
		    if (USEX(SCI_VOL) != 0xffffU)
		      USEX(SCI_VOL) += 0x0101;
		setvol:
#ifdef SAVE_VOLUME
		    volChanged = 1;
#endif/*SAVE_VOLUME*/
#ifdef USE_PRINTABLE_OUTPUT
		    putdec(USEX(SCI_VOL) & 255);
		    putch('\n');
#else
		    putch(USEX(SCI_VOL));
#endif
                    #endif/*elseUART_BUFFERED*/
		} else if (ui.Key == SW3) {

		  if (ui.KeyTime == 12) { /* SW3 long press */
		      ui.KeyTime = -32767; /*one-shot*/
		      if (cachedSpiState == spi_waitima) {
			/* If in record mode, end record mode. */
			goto endRec;
		      }
		      /* goto record if not already */
		      USEX(SCI_AICTRL3) |= CTRL3_RECORD_ON;
		  }
		}
	    }
	} else { // PKP: Keystate is not the same as previous keystate
	  // PKP: Handle short keypresses
	  if (ui.Data == 0 && ui.KeyTime > 0) {
	    // PKP: Handle end of short keypress (current keystate ui.Data=0)
	    /* short presses */
	    if (ui.KeyTime < 12) {
	      /* was pressed < 0.5 seconds */
	      
	      if (ui.Key == SW3) {
		/* play pause/continue - works for encoding too. */
		USEX(SCI_AICTRL3) ^= CTRL3_PAUSE_ON;
	      } else if (ui.Key == SW1 || ui.Key == SW2) {
		if (ui.Key == SW1) {
		  #ifdef UART_BUFFERED
		  ui.Key = ui.Data;
		  goto nextsong;
                  #else
		  USEX(SCI_AICTRL0) += 0x8001;
                  #endif
		} else {
		  #ifdef UART_BUFFERED
		  ui.Key = ui.Data;
		  goto prevsong;
		  #else
		  if (USEX(SCI_DECODE_TIME) < 4) {
		      if (USEX(SCI_AICTRL0) == 0) {
			  USEX(SCI_AICTRL0) = USEX(SCI_AICTRL1) + (0x8000-1);
		      } else {
			  USEX(SCI_AICTRL0) += (0x8000-1);
		      }
		  } else {
		      USEX(SCI_AICTRL0) |= 0x8000U;
		  }
                  #endif
		}
	      }
	    }//(ui.keyTime was < 12)
	  }
	  ui.Key = ui.Data;
	  ui.KeyTime = 0;
	}
        
#if defined(SAVE_VOLUME)
	/*Save volume when volume button is released. */
	if (ui.Data == 0 && volChanged && cachedSpiState == spi_seek) {
	  volChanged = 0;
	  SaveWord(26/*volume=26*/, USEX(SCI_VOL));
	}
#endif/*SAVE_VOLUME*/
	//ui.Data = 0;
    }
#endif/*HAS_KEYS*/



#ifdef UART_BUFFERED /* This is becoming the new default. */
    if (UartFill() != 0) {
      int c = UartGetByte();
      if (c == '.') { /* next */
      nextsong:
	USEX(SCI_AICTRL0) += 0x8001;
      } else if (c == ',') { /* previous */
      prevsong:
	if (USEX(SCI_DECODE_TIME) >= 4) {
	  goto cancel;//USEX(SCI_AICTRL0) |= 0x8000U; //restart the same file
	}
	USEX(SCI_AICTRL0) = USEX((USEX(SCI_AICTRL0) == 0) ? SCI_AICTRL1 : SCI_AICTRL0) + (0x8000-1);
      } else if (c == 'C') { /* Cancel */
      cancel:
	USEX(SCI_AICTRL0) |= 0x8000;
	if (cachedSpiState == spi_waitima) {
	  /* Cancel record and prevent saving file by making size 0. */
	  spix.size = 0;
	  //putch('C');
	}
#if !defined(USE_DASH_IN_RECORDER) || defined(NO_ENCODE_MP3)
      } else if (c == 0xef) {
	USEX(INT_ENABLE) = 0;
	Monitor(); //to allow easy reprogramming
#endif
#ifdef USE_EARSPEAKER
      } else if (c == 'e') {
	parametric_x.earSpeakerLevel ^= 0x6000; /*EarSpeaker toggle*/
#endif
#ifdef USE_FILEMODE
      } else if (c == 'f') { /* filemode */
	fileMode = 1;
#endif
#if !defined(NO_RANDOM) || defined(USE_FILEMODE)
      } else if (c == 'c') { /* continuous play mode - random off */
#ifndef NO_RANDOM
	USEX(SCI_AICTRL3) &= ~CTRL3_RANDOM_PLAY;
#endif
#ifdef USE_FILEMODE
	fileMode = 0;
#endif
#endif/*!defined(NO_RANDOM) || defined(USE_FILEMODE)*/
#ifndef NO_RANDOM
      } else if (c == 'r') { /* random play */
	USEX(SCI_AICTRL3) |= CTRL3_RANDOM_PLAY;
#endif
#ifndef NO_LOOP_MODE
      } else if (c == 'l') { //25 words
	register u_int16 ai3 = USEX(SCI_AICTRL3);
	if ((ai3 & CTRL3_PLAY_MODE_MASK) == CTRL3_LOOP_SONG) {
	  USEX(SCI_AICTRL3) = ai3 & ~CTRL3_PLAY_MODE_MASK;
	  putch('>');
	} else {
	  USEX(SCI_AICTRL3) = (ai3 & ~CTRL3_PLAY_MODE_MASK) | CTRL3_LOOP_SONG;
	  putch('l');
	}
#endif
      } else if (c == 'R') { /* Record! */
	USEX(SCI_AICTRL3) |= CTRL3_RECORD_ON;
      } else if (c == 'E') { /* End Record! */
	goto endRec;
	
      } else if (c == '=') { /* pause */
#ifdef USE_PRINTABLE_OUTPUT
	putstrp("\ppause\n");
#else
	putch('=');
#endif
	USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
      } else if (c == '>') { /* play 1x */
	USEX(SCI_AICTRL3) &= ~CTRL3_PAUSE_ON;
	parametric_x.playSpeed = 1;
#ifdef USE_PRINTABLE_OUTPUT
	putstrp("\pplay\n");
#else
	putch('>');
#endif
      } else if (c == 0xbb /* '»' */) {
	/* play faster */
	parametric_x.playSpeed++;
#ifdef USE_PRINTABLE_OUTPUT
	putdec(parametric_x.playSpeed);
	putch('\n');
#else
	putch(parametric_x.playSpeed);
#endif
      } else if (c == '+') { /* volume up */
      volup:
	if (USEX(SCI_VOL))
	  USEX(SCI_VOL) -= 0x0101;
	goto setvol;
      } else if (c == '-') { /* volume down */
      voldown:
	if (USEX(SCI_VOL) != 0xffffU)
	  USEX(SCI_VOL) += 0x0101;
      setvol:
#ifdef SAVE_VOLUME
	volChanged = 1;
#endif/*SAVE_VOLUME*/
#ifdef USE_PRINTABLE_OUTPUT
	putdec(USEX(SCI_VOL) & 255);
	putch('\n');
#else
	putch(USEX(SCI_VOL));
#endif
      } else if (c == '?') { /* decode time + info -- Note: will block.*/
#ifdef USE_PRINTABLE_OUTPUT
	putdec(USEX(SCI_DECODE_TIME));
	putch(' ');
	/* play file position from 0..255 */
	putdec((fatInfo.fileSize-2*spix.size) / ((fatInfo.fileSize >> 8)+1));
	putch('\n');
#else
	putword(USEX(SCI_DECODE_TIME));
	/* play file position from 0..255 */
	putch((fatInfo.fileSize-2*spix.size) / ((fatInfo.fileSize >> 8)+1));
#endif
      } else {
	/* unknown command */
	//putch(c);
      }
    }
#endif/*UART_BUFFERED*/


    /* The following now handles the (file-related) operations for:
       1) the UART control
       2) the SCI control
       3) the button control
    */


    /* play pause/continue for both playback and recording */
    if ((USEX(SCI_AICTRL3) & CTRL3_PAUSE_ON)) {
      /* CTRL3_PAUSE_ON overlaps the record pause bit. */
      parametric_x.playMode |= PLAYMODE_PAUSE_ON;
    } else {
      parametric_x.playMode &= ~PLAYMODE_PAUSE_ON;
    }
    /* Check recording... */
    if ((USEX(SCI_AICTRL3) & CTRL3_RECORD_ON)) {
      if (cachedSpiState != spi_waitima && cachedSpiState != spi_record &&
	  cachedSpiState != spi_zeros && record.record != 0x2468) {
	/* If not already recording, start recording */
	record.record = 0x1234; /* request rec */
	//putstrp("\pStart Rec!\n");
	goto out_of_wav;
      }
    } else {
    endRec:
      if (cachedSpiState == spi_waitima) {
	/* If recording, end recording */
	USEX(SCI_AICTRL0) = 0;
	//putstrp("\pEnd Rec!\n");
	/* End recording nicely to create valid files. */
	record.record = 1;
	USEX(SCI_AICTRL3) &= ~AICTRL3_PAUSEON; /*automatically remove pause*/
	USEX(SCI_MODE) |= (1<<SCIMB_CANCEL); /*flag encoder to finalize*/
	//putch('1');
      }
    }


    if ((USEX(SCI_AICTRL0) & 0x8000U)) {    /* new song */
      USEX(SCI_AICTRL0) &= 0x7fff; /* ack command */
      USEX(SCI_AICTRL3) &= ~CTRL3_RECORD_ON; /* end recording */
    out_of_wav:
      USEX(SCI_MODE) |= (1<<SCIMB_CANCEL);
      csCancel = 1;
      parametric_x.resync = 0; /* disable resyncs */

#ifdef USE_READ_MULTIPLE_BLOCK_ONE
      MmcStopTransmission();
      SpiSendClocks();
#else
      /* Discard any unread data, including CRC, before starting again */
      {
	register int i;
	for (i=0; i<256+2; i++)
	  SpiSendReceiveMmc(-1/*0xffff*/,16);
      }
      SpiSendClocks();
#endif

      /* If we were recording */
      if (cachedSpiState == spi_waitima) {
#ifdef UART_OUTPUT
#else /*UART_OUTPUT*/
	/* wait until MMC write finished */
	spiCnt = 0;
	while ( SpiSendReceiveMmc(-1/*0xffffU*/, 16) != 0xffffU) {
	  if (++spiCnt >= 32000)
	    break; /* timeout */
	}
	SpiSendClocks();
	Disable(); //prevent ADC from trashing the FAT
	/* In vs1063 the stream buffer is used as SRC_BUFFER. */

	if (spiCnt < 32000 && (mmc.fileSize = spix.size)) {
	  
	  /* If finalized correctly and we have saved anything -
	     if we cancel recording, spix.size will be set to 0.

	     Set the suffix. The suffix can be decided here because
	     FatFindDirEntry ignores the suffix, so the name will be
	     unique regardless of the suffix. */
	  {
	    register u_int16 fmt = (USEX(SCI_AICTRL3) & 0x00f0);
#ifndef NO_ENCODE_MP3
	    if (fmt == AICTRL3_MP3) {
	      mmc.fileName[4] = ('M'<<8) | 'P';
	      mmc.fileName[5] = ('3'<<8);
	    } else
#endif
	    if (fmt == AICTRL3_VORBIS) {
	      mmc.fileName[4] = ('O'<<8) | 'G';
	      mmc.fileName[5] = ('G'<<8);
	    } else {
	      mmc.fileName[4] = ('W'<<8) | 'A';
	      mmc.fileName[5] = ('V'<<8);
	      /* Fix WAV size fields, but only for WAV! */
	      ReadDiskSectorTo(mmc.freeStart, stream_buffer);
	      {
		register u_int32 t = mmc.fileSize - 8;
		stream_buffer[2] = SwapWord(t);
		stream_buffer[3] = SwapWord(t>>16);
		t -= 0x28;
		stream_buffer[22] = SwapWord(t);
		stream_buffer[23] = SwapWord(t>>16);
	      }
	      /* We don't care if write fails or not. */
	      MmcWrite(mmc.freeStart, stream_buffer);
	    }
	  }
	  //putstrp("\pMaking file");
	  /* FsFatMakeFile() creates the directory entry and allocates
	     the data blocks. */
	  FsFatMakeFile();

	  /* Full reset required before or after
	     encoding for it to work more than once.

	     Well, it's better after because the encoding
	     patch should never return anyway.
	  */
	  putstrp("\pSaving Recording!\n");
	} else {
	  putstrp("\pCancel Recording!\n");
	}
#endif /*elseUART_OUTPUT*/
	      
	/* Use wdog-resert to prevent DECIM-restart-bug (channel swap)
	   in our next recording start.
	   Watchdog reset will also clear the recording indicator.
	*/
	clockX = 2;
	SetRate(10); /* Need 1x clock for watchdog to work properly! */
	USEX(WDOG_CONFIG) = 2;
	USEX(WDOG_RESET) = WDOG_RESET_VAL;
	while (1)
	  ;
      }

      /* Clear pause and file ready bits automatically */
      cachedSpiState = spi_zeros; /* continue with play cancel procedure */
      USEX(SCI_AICTRL3) &= ~(CTRL3_PAUSE_ON|CTRL3_FILE_READY|CTRL3_AT_END);
      spiCnt = 0;
    }
    

    if (cachedSpiState == spi_waitima) {
      /* spi_waitima: We are waiting for new data to arrive from the encoder
	 and write data blocks when there is data available and the card
	 is ready for write. */
      if (record.record == 3 && record.fill <= 0) {
	/* We have written everything. Now finalize the file. */
	putstrp("\pFinish Rec!\n");
	goto out_of_wav;
      }
#ifdef UART_OUTPUT
      
#else/*elseUART_OUTPUT*/
      /* Check if SD has the response for us. */
      if (SpiSendReceiveMmc(-1/*0xffffU*/, 16) == 0xffffU) {
	/* SD Ready. */
	SpiSendClocks(); // extra clocks with chip select high
	SpiSendClocks(); // extra clocks with chip select high
	    
	/* Note: byteswap is not required with vs1063a.

	   Because SD cards take longer time to write sectors
	   when they switch from one erase block to another,
	   we need quite a long FIFO for the audio data.
	   The FIFO in the encoding routine is not enough,
	   so we 'receive' data from the encoder's FIFO and
	   store it into our bigger FIFO and save it from there.

	   Note: could use MMC_WRITE_MULTIPLE_BLOCK...
	*/
	/* Data in second FIFO or last sector to write. */
	if (record.fill >= 512 || record.record == 3) {
	  /* Save a sector from the second FIFO */
	  //putch('W');
	  WritePhysicalSectorY(spix.sector<<mmc.hcShift,
			       (__mem_y void *)(YBUF+record.rd));
	  //putch('M');
  
	  spix.sector++;
#if 0
	  if (record.fill < 512) { /*last sector*/
	    spix.size += record.fill; /* in bytes! */
	  } else {
	    spix.size += 512; /* in bytes! */
	  }
#else
	  spix.size += (record.fill < 512) ? record.fill : 512;
#endif
	  record.fill -= 512; //can go to negative and still end recording.
	  record.rd += 256;
	  if (record.rd >= YBUFSZ) {
	    record.rd -= YBUFSZ;
	    //putch('<');
	  }
	  //putch('.');
	}
	/* Limit filesize to about 2GB or to the available space. */
	if (spix.size >= 0x7ffff000 || /*max size about 1 day at 24kHz*/
	    spix.size/512 >= mmc.freeSize ) {
	  /* disk full, end recording. */
	  putstrp("\pdisk full\n");
	  goto endRec;
 	}
      } else {
	//putch('.');
      }
      USEX(SER_DREQ) = 0;
  
      /* Fill second FIFO from encoder's FIFO. */
#define READ_SIZE 32 /*in words -- must be 32, 64, or 128 */
      if (record.fill < YBUFSZ - READ_SIZE) {
	register s_int16 imaFil = (bssWp - bssRp);
	__mem_y s_int16 read_size = READ_SIZE;
	if (imaFil < 0)
	  imaFil += bssSize;
	//putword(imaFil);
	//putstrp("\pREAD_SIZE\n");
	if (record.record == 1 &&
	    (USEX(SCI_MODE) & (1<<SCIMB_CANCEL)) == 0) {
	  if (imaFil < read_size) {
	    read_size = imaFil;
	    record.record = 2;
	    //putch('2');
	  }
	}

	if (imaFil && imaFil >= read_size) {
	  register s_int16 newend = record.wr + read_size;
	  if (newend >= YBUFSZ) 
	    newend -= YBUFSZ;
	  //putstrp("\pIMAFIL\n");
	  //if (newend != record.rd)
	  {
	    //putch('r');
	    memcpyYY(YBUF+record.wr, bssRp, read_size);
	    bssRp += read_size;
	    if (bssRp >= bssEnd) {
	      bssRp -= bssSize;
	    }
	    record.wr = newend;
	    record.fill += 2*read_size;
	  }
	  /* In vs1063 encoder uses WmaStereoCopy()
	     but in codec mode only, so normally
	     applAddr is not called to provide us a
	     timebase. We use a timer for timebase now. */
	}
	if (record.record == 2) {
	  if ((s_int16)parametric_x.endFillByte < 0) {
	    *(__y u_int16 *)(YBUF+record.wr) = parametric_x.endFillByte << 8;
	    record.fill += 1;
	    parametric_x.endFillByte = 0;
	  }
	  record.record = 3;
	  //putch('3');
	}
      }
#endif/*elseUART_OUTPUT*/
      
#if defined(LOWER_LOOPBACK_VOL) /* lower loopback volume */
      volume[0] = LOWER_LOOPBACK_VOL;
      volume[1] = LOWER_LOOPBACK_VOL;
#endif
    } else if (cachedSpiState == spi_record) {
#ifdef HAS_KEYS
      //Encoding overrides interrupt enable
      USEX(INT_ENABLE) |= (1<<INT_EN_TIM0);
#endif
#if 0
      putstrp("\pspi_rec\n");
#endif
      record.fill = record.wr = 0;
      spix.size = 0; /* calculate file size, in bytes! */
#ifndef UART_OUTPUT
      /* The WAV header is part of the data in vs1063. */
      spix.sector = mmc.freeStart;
      cachedSpiState = spi_waitima;
#endif
    }

    if (cachedSpiState == spi_file) { /* Play file from the beginning. */
        /* extra byte will be read if the file size is odd */
        spix.size = (fatInfo.fileSize+1)/2; /* convert to words */
        cachedSpiState = spi_seek;
	//setfrag:
        spix.fragment = &fragments[0];
        spix.sector   = fragments[0].start /*& 0x7fffffffUL*/;
        spix.fragSize = fragments[0].size; /*todo: *fatInfo.sectorsPerCluster*/
    }
    //checkseek:
    if (cachedSpiState == spi_seek) {
#ifdef USE_READ_MULTIPLE_BLOCK_ONE
	MmcCommand(MMC_READ_MULTIPLE_BLOCK|0x40,
	  (spix.sector & 0x7fffffffUL) << mmc.hcShift);
#else
	/* Send the SD/MMC read block command */
	MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,
		   (spix.sector & 0x7fffffffUL) << mmc.hcShift);
#endif /*elseUSE_READ_MULTIPLE_BLOCK_ONE*/
        cachedSpiState = spi_waitdata;
        spiCnt = 0;
    }

    /*
      We have sent the read block command and now check if the data
      is available to be read out.
     */
    if (cachedSpiState == spi_waitdata) {
	register u_int16 res = SpiSendReceiveMmc(0xff00,8);
	if (res == 0xfe) { /* 0xfe is the data start token */
	    cachedSpiState = spi_data; /* go to fetch mode */
	    spiCnt = 0;
	    NextSector();
	} else if (res == 0xff) { /* card is still fetching block */
	    if ((u_int16)++spiCnt > 50000) {
	      //putstrp("\pspi_waitdata timeout\n");
	      MyReset(); /* timeout, reset immediately */
	    }
	} else {
	    /* Otherwise we have Read Error, and data is not transferred */
	  //putword(res);
	  //putstrp("\p Read Error\n");
	    goto end_of_file;
	}
    } else if (cachedSpiState == spi_data) { /* data transfer state */
	/* Loop until the whole file is read, or stream buffer is full,
	   or the next sector needs to be read. */
	while (spix.size > 0) {
	    register int i, j = 32;
	    register u_int16 *wr;
	    if (USEX(SCI_HDAT1) == (('f'<<8) | 'L')) {
	      register s_int16 t = stream_wr_pointer - stream_rd_pointer;
	      if (t < 0)
		t += 0x1800;
	      if (t > 0x1800/*STREAM_BUFFER_SZ*/-40)
		break; /* stream buffer full, leave loop */
	      wr = (void *)stream_wr_pointer;
	      if (j > spix.size)
		j = spix.size;
	      for (i=0;i<j;i++) {
		*wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
		if (wr >= &stream_buffer[0x1800/*STREAM_BUFFER_SZ*/])
		  wr = stream_buffer;
	      }
	    } else {
		if (StreamDiff() >= STREAM_BUFFER_SZ-40)
		    break; /* stream buffer full, leave loop */
		wr = (void *)stream_wr_pointer;
		if (j > spix.size)
		    j = spix.size;
		for (i=0;i<j;i++) {
		    *wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
		    if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
			wr = stream_buffer;
		}
	    }
	    stream_wr_pointer = (void *)wr;
	    spix.size -= j;//32;
	    spiCnt += 32;
	    if (spiCnt >= 256) {
		/* 256 words read, get next block in file */
		SpiSendReceiveMmc(-1/*0xffff*/,16); /* discard crc */
		cachedSpiState = spi_seek;
#ifdef USE_READ_MULTIPLE_BLOCK_ONE
	        MmcStopTransmission();
#endif
		break;
	    }
	    /* watchdog clearing here doesn't work with midi */
	}
  
        if (spix.size <= 0) {
end_of_file:
#ifdef USE_READ_MULTIPLE_BLOCK_ONE
	  MmcStopTransmission(); /* Stop Transmission */
	  SpiSendClocks();
#else
	  /* Discard any unread data, including CRC, before starting again */
	  {
	    register int i;
	    for (i=0; i<256+2; i++)
	      SpiSendReceiveMmc(-1/*0xffff*/,16);
	  }
	  SpiSendClocks();
#endif
	  {
	    register int m = (USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK);
	    if (m == CTRL3_LOOP_SONG){
	      /* restart the same file immediately */
	      cachedSpiState = spi_file;
	    } else {
	      /*NEW: pause after play */
	      if (m == CTRL3_PAUSE_AFTER_PLAY) {
		USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON | CTRL3_AT_END;
	      } else {
		USEX(SCI_AICTRL0) += 1; /* next file */
	      }
	      cachedSpiState = spi_zeros;
	      //putch('Z');
	      spiCnt = 0;
	      if (USEX(SCI_HDAT1) == (('f'<<8) | 'L'))
		USEX(SCI_MODE) |= (1<<SCIMB_CANCEL);
	    }
	  }
        }
#ifdef NO_WMA /* reject WMA files */
	if (USEX(SCI_HDAT1) == (('W'<<8)|'M')) {
	    USEX(SCI_AICTRL0) += 1; /* next file */
	    goto goreset;//MyReset();
	}
#endif
#ifdef NO_AAC /* reject AAC files */
	{
	    register u_int16 t = USEX(SCI_HDAT1);
	    //0x4154(AT) 0x4144(AD) 0x4d34(M4)
	    if (t == 0x4154 || t == 0x4144 || t == 0x4d34) {
		USEX(SCI_AICTRL0) += 1; /* next file */
		goto goreset;//MyReset();
	    }
	}
#endif

    } else if (cachedSpiState == spi_zeros) { /* send zeros after the file */
#ifdef USE_READ_MULTIPLE_BLOCK_ONE
      //
#else
	if (spiCnt == 0) {
            /* Discard any unread data, including CRC, before starting again */
	    register int i;
	    for (i=0; i<256+2; i++)
		SpiSendReceiveMmc(-1/*0xffff*/,16);
	    SpiSendClocks();
	}
#endif

        /* Workaround for WMA resync after stream ended! */
        parametric_x.resync = 0;

	/*TODO: different for flac */
#if 0//no space in recorder
	/* FLAC files use a bigger stream buffer,
	   also requires 12k of endFillByte! */
	if (USEX(SCI_HDAT1) == (('f'<<8) | 'L')) {
	    register s_int16 t = stream_wr_pointer - stream_rd_pointer;
	    if (t < 0)
		t += 0x1800;
	    if (t < 0x1800/*STREAM_BUFFER_SZ*/-40) {
		register int i, z = parametric_x.endFillByte * 0x0101;
		register u_int16 *wr = (void *)stream_wr_pointer;
		for (i=0;i<32;i++) {
		    *wr++ = z;
		    if (wr >= &stream_buffer[0x1800/*STREAM_BUFFER_SZ*/])
			wr = stream_buffer;
		    ++spiCnt;
		    stream_wr_pointer = (void *)wr;
		}
	    }
	} else
#endif
	  {
	  if (StreamDiff() < STREAM_BUFFER_SZ-40) {
            register int i, z = parametric_x.endFillByte * 0x0101;
            register u_int16 *wr = (void *)stream_wr_pointer;
//	    if (USEX(SCI_MODE) & (1<<SCIMB_CANCEL)) {
//		z = 0xfffa;
//	    }
            for (i=0;i<32;i++) {
                *wr = z;
                if (wr >= &stream_buffer[STREAM_BUFFER_SZ-1]) {
                    wr = stream_buffer;
                } else {
                    wr++;
                }
                ++spiCnt;
                stream_wr_pointer = (void *)wr;
            }
            /* watchdog clearing here doesn't work with midi */
	  }
	}

	if (spiCnt >= 256+1*1024) {
	    /* Cancel play at the end of the song. */
	    USEX(SCI_MODE) |= (1<<SCIMB_CANCEL);
	    csCancel = 1;
	}
	/* If cancel didn't seem to do it, then restart. */
	if (USEX(SCI_HDAT1) == (('f'<<8) | 'L') && spiCnt >= 256+6*1024) {
	    goto goreset;
	} else if (//USEX(SCI_HDAT1) == 0 ||
		   //(USEX(SCI_HDAT1) == 0x7665 /*"ve"*/ && spiCnt >= 256+1024+2*64)  ||
		   spiCnt >= 256+4*2048) {
	    /* 2050 bytes */
goreset:
	    putstrp("\pcancel too slow!\n");
//putch('e');
	    MyReset();
	}
    }
    

#ifdef HAS_KEYS
#ifdef NO_DREQ_LED
    /* Save space by not updating DREQ. */
#else /*NO_DREQ_LED*/
    /* 15 words */
    {
      register int t = USEX(SCI_AICTRL3) & 1;  /* lowest bit is global random play*/
      /* DREQ indicates random play status */
      /*
	Play:       XXRR............ 0x1c!=0 -> 0 0x1c==0->1
	Play Pause: XR......XR...... 0x0e!=0 -> 0 0x0e==0->1
	RecPause    XXXX....XXXX.... 0x0e!=0 -> 0, otherwise 1
	Record      XXXX............ 0x1c!=0 -> 0, otherwise 1
      */
      u_int16 mask = 0x1f;
      if ((USEX(SCI_AICTRL3) & AICTRL3_PAUSEON)
	  || cachedSpiState == spi_pause) {
	mask = 0x3f;
      }
      mask = ui.Timer & mask; /*ui.Timer only available with keys.*/
      if (mask <= 4) {
	t = 1;
      } else if (cachedSpiState == spi_waitima || mask >= 8) {
	t = 0;
      }
      USEX(SER_DREQ) = t;
    }
#endif/*!NO_DREQ_LED*/
#endif/*HAS_KEYS*/
    spiState = cachedSpiState;
}



void NewSinTest(void);



#if 0 //Bring a default version from the library
void IterateFilesCallback(register __b0 u_int16 *name) {
    register int i;
    /* Dump the whole FAT directory entry */
    for (i=0;i<(32/2);i++) {
        putword(*name++);
    }
}
#endif


void flacStreamBufferRestore(void);
auto void MyMain(register s_int16 __a1 fullReset) {
    /* Clear the hooks if started from SCI */
    applAddr = NULL;
    USEX(SCI_AIADDR) = 0;

#ifdef ENABLE_HIGHREF
    USEX(SCI_STATUS) |= (1<<SCIST_REFERENCE_SEL); /*1.65V*/
#endif
    /* Install new idle hook handler */
    WriteIMem((u_int16)IdleHook, 0x28000000UL | ((u_int32)(u_int16)UserHook<<6));
  #ifdef UART_BUFFERED
    (void)UartFill(); /* pull in UART stuff from the link library */
  #endif

    /* Make sure we have the right upper ROM visible. */
    USEX(VS1053_IROM4) &= ~(IROM4_ENABLE);
    /* Just in case, restore the normal SDI interrupt handler (1024 words). */
    flacStreamBufferRestore();
    USEX(SER_DREQ) = USEX(SCI_AICTRL1) = USEX(SCI_HDAT1) = 0; /* clear so FLAC is not indicated.*/
    /* clear mp4ASC, seems to be needed after FLAC. */
    memset((void *)0x2000, 0, 64); /*Note: different addr for vs1053 and 63*/

    /* For MMC/SD handling the SCI_NEWMODE must be set, and also
       the relevant GPIO pins configured correctly. */

    USEX(SCI_MODE) = (USEX(SCI_MODE) & ~(1<<SCIMB_CANCEL)) |
      ((1<<SCIMB_SDI_NEWMODE) | (1<<SCIMB_LINE)
#ifdef ENABLE_LAYER12
       | (1<<SCIMB_ALLOW_LAYER12)
#endif
       );
    USEX(GPIO_ODATA) = SPI_xCS|SPI_CLK|SPI_xCS2|SD_POWER_ENABLE;
    USEX(GPIO_DDR)   = SPI_xCS|SPI_CLK|SPI_xCS2|SD_POWER_ENABLE
    #ifdef RECORDING_INDICATOR
	| RECORDING_INDICATOR
    #endif
    #ifdef PREPARE_RECORDING_INDICATOR
	| PREPARE_RECORDING_INDICATOR
    #endif
    #if defined(ENABLE_I2S)
	/* Enable I2S pins GPIO4-GPIO6 */
	| 0xf0;
      USEX(I2S_CONTROL) = 0x0c
    #endif
	;

    /* InitHardware() configures clock/PLL according to SCI_CLOCKF,
       initializes stream and audio buffer pointers, sets default
       samplerate (8000Hz), clears HDAT0,HDAT1,DECODE_TIME,AUDATA,
       and sets the cosmetic chip ID in parametric_x.chipID .

       User should set uartByteSpeed 960 before the call to get the
       default 9600bps transfer rate. (vs1063)
    */
    hwSampleRate = 11; /*force, mem is only cleared at boot, not restart*/
//    uartByteSpeed = 9600/10; //for vs1063
    uartByteSpeed = 115200/10; //for vs1063
    InitHardware(); /*no longer clears memory in vs1053/63 so preserves const*/
#if 0
    audioPtr.curFctl = audioPtr.newFctl = 0x15555; //correct for 8kHz 20111117
#else
    SetRate(8001);
#endif

  #ifndef USE_DEBUG
    /* Enable UART RX interrupt to be able to jump to emulator.
       Allows reflashing without MMC/SD inserted. */
    USEX(INT_ENABLE) = (1<<INT_EN_RX);
  #endif

    /* If not powerup reset, re-enable analog drivers immediately.
       InitHardware() doesn't actually touch powerdowns.
       CheckVolumeAndBass() clears analog powerdown.
     */
    if (!fullReset) {
      //USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN2);
    }

#ifdef HAS_KEYS
    USEX(TIMER_CONFIG) = clockX;
    *(volatile u_int32 *)(TIMER_T0L) = 6000000/32; /*16Hz->32Hz*/
    USEX(TIMER_ENABLE) |= 1;
    USEX(INT_ENABLE) |= (1<<INT_EN_TIM0);
#endif

    /* Enable interrupts */
    USEX(INT_GLOB_ENA) =
#ifdef USE_DASH_IN_RECORDER
      (u_int16)AacPatch /*Pull in DASH support for AAC. */
#else
      0
#endif
      ;
    USEX(INT_GLOB_ENA) =
#ifdef HAS_KEYS
      (u_int16)TimerInt /*Pull in timer interrupt - provides a timebase.*/
#else
      0
#endif
      ;
    //USEX(INT_GLOB_ENA) = 0; //two should be enough

    /* wake up MMC! -- If this takes too long, watchdog will bite. */

  #ifdef SAVE_POSITION
    /* Clear the current song number so it will reset between cards. */
    if (USEX(SCI_AICTRL0)) /* 9 words */
	SaveWord(28/*AICTRL0*/, 0); //3 words
  #endif


#ifdef UART_BUFFERED
    //putstrp("\pInit MMC\n");
#endif
    memsetY(&mmc, 0, sizeof(mmc));
 cardchanged:
    memset(&fatInfo, 0, sizeof(fatInfo));
    InitializeMmcRecorder(50); //if the same card re-inserted, use cached values
    if (mmc.state != mmcOk || mmc.errors) {
      putword(('!'<<8)|'C');//putstrp("\p!C");
      goto cardchanged;
    }
#ifdef USE_PRINTABLE_OUTPUT
    putstrp("\pCard ok\n");
#endif
    // Init File System (FAT)
    /* Detects cards without partition table */
    memset(stream_buffer, 0, 512/2);
    if ( InitFileSystem() != 0) { //0x55ab - found 0 instead of 0xaa55
#ifdef UART_BUFFERED
#if 0
      register u_int16 *p = stream_buffer;
      register int i;
      for (i=0;i<256;i++)
	puthex(*p++);
#endif
      putstrp("\p\nNo FAT\n");
#endif
	goto cardchanged;
    }
    //puthex(fatInfo.FilSysType);
    putstrp("\pFAT ok\n");
    
  #ifdef HAS_KEYS
    ui.Key = 0;
    ui.KeyTime = -32767; /*Ignore record button release*/
  #endif

#ifdef CTRL3_NO_NUMFILES
    if ((USEX(SCI_AICTRL3) & CTRL3_NO_NUMFILES)) {//+7 words
      /* Startup is faster if we don't need to find number of files. */
      USEX(SCI_AICTRL1) = 0x7fff;
    } else
#endif
    {
      /* determine number of playable files */
      USEX(SCI_AICTRL1) = OpenFile(0xffffU);
    }
#ifdef USE_PRINTABLE_OUTPUT
    /* In decimal format */
    putstrp("\pfat ");
    putdec(fatInfo.totSize);
    putstrp("\p\nfiles ");
    putdec(USEX(SCI_AICTRL1));
    putch('\n');
#else
    #if 0
    /* In hex format */
    putstrp("\pfat ");
    puthex(fatInfo.totSize>>16);
    puthex(fatInfo.totSize);
    putstrp("\p\nfiles ");
    puthex(USEX(SCI_AICTRL1));
    putch('\n');
    #endif
    putword(USEX(SCI_AICTRL1));
    putch('\n');
#endif



 fileloop:
#ifdef USE_FILEMODE
    if (fileMode) {
	static u_int16 buffer[80];
	register u_int16 *bf = buffer;

	//USEX(SCI_AICTRL0) = -1;
	goto prompt;
	while (1) {
	    if (UartFill()) {
	      __mem_y int cmd = UartGetByte();
		*bf = 0;
//putch(cmd);
		if (cmd == '\n' || bf == &buffer[79]) {
		    /* Build a 8.3-character filename */
#if 0
		    fileName[0] = (buffer[1]<<8) | buffer[2];
		    fileName[1] = (buffer[3]<<8) | buffer[4];
		    fileName[2] = (buffer[5]<<8) | buffer[6];
		    fileName[3] = (buffer[7]<<8) | buffer[8];
		    fileName[4] = (buffer[9]<<8) | buffer[10];
		    fileName[5] = (buffer[11]<<8) | 0x20;
		    /* Set the Archive flag attribute for files we create. */
#else
		    fileName[0] = buffer[2] + buffer[1]*256;
		    fileName[1] = buffer[4] + buffer[3]*256;
		    fileName[2] = buffer[6] + buffer[5]*256;
		    fileName[3] = buffer[8] + buffer[7]*256;
		    fileName[4] = buffer[10] + buffer[9]*256;
		    fileName[5] = 0x20 + buffer[11]*256;
		    /* Set the Archive flag attribute for files we create. */
#endif

		    /* Whenever a command is entered,
		       poll MMC/SD present by giving it a command.
		       But only when idle, i.e. mmcOk. */
		    if (mmc.state == mmcOk && mmc.errors == 0 &&
			MmcCommand(MMC_SET_BLOCKLEN|0x40, 512) != 0) {
			mmc.errors++;
#if defined(USE_PRINTABLE_OUTPUT) || defined(UART_BUFFERED)
			putstrp("\pcard removed\n");
#endif
			goto cardchanged;
		    }
		    switch (buffer[0]) {
#if 1
		    case 'S': /* Set recording time and date */
			/*Date bits
			  15-9 	Year (0 = 1980, 127 = 2107)
			  8-5 	Month (1 = January, 12 = December)
			  4-0 	Day (1 - 31)
			  2015: 35*512 = 17920
			  Oct:  10 * 32 = 320
			  22:   +22
			  2015-Oct-22 = 18262
			*/
			created[0] = atou((void *)&buffer[1]);
			bf = &buffer[1];
#ifndef HAS_KEYS //no space for this right now if both UART and buttons
			while (*bf == ' ')
			    bf++;
			while (*bf && *bf != ' ')
			    bf++;
			if (*bf) {
			    /*Time bits
			      15-11	Hours (0-23)
			      10-5 	Minutes (0-59)
			      4-0 	Seconds/2 (0-29)*/
			    created[1] = atou((void *)bf);
			}
#else
			//created[1] = 0; //set no tod - initialized to 0
#endif
			//created[2] = 0; //10ms units from 0 to 199 - init 0
#ifdef USE_PRINTABLE_OUTPUT
			putdec(created[0]);
			putch(' ');
			putdec(created[1]);
			//putch(' ');
			//putdec(created[2]);
			putch('\n');
#else
			//putword(created[0]);
			//putword(created[1]);
#endif
			break;
#endif

		    case 'c':
			/* continuous play mode */
			fileMode = 0;
			goto playloop;

		    case 'L':
			/* List files - calls IterateFilesCallback() */
			USEX(SCI_AICTRL3) |= CTRL3_ITERATE;
			OpenFile(0x7fff);
			//USEX(SCI_AICTRL3) &= ~CTRL3_ITERATE; /*turned off by OpenFile() */
			break;

		    case 'R':
			record.record = 0x9234; /* request rec */
			goto recstart;

#ifdef DELETE_FILE
		    case 'D':  /* 'D'elete by name */
			/* already put the name into the fileName variable */
			fatInfo.dirLine = -1;
			USEX(SCI_AICTRL3) |= CTRL3_BY_NAME;
			OpenFile(0x7fff); /*OpenFile updates AICTRL0*/
			//USEX(SCI_AICTRL3) &= ~CTRL3_BY_NAME; //Cleared by OpenFile()
			if (fatInfo.dirLine >= 0) {
#ifdef USE_PRINTABLE_OUTPUT
			    putstrp("\pDeleting\n");
#endif
			    /* FsDeleteFile() requires dirSector, dirLine,
			       and fragments[] initialized. */
			    FsDeleteFile();
			} else {
#ifdef USE_PRINTABLE_OUTPUT
			    putstrp("\pnot found\n");
#else
			    putstrp("\pnof\n");
#endif
			}
			break;
#endif

		    case 'P':  /* 'P'lay by name */
			/* already put the name into the fileName variable */
			USEX(SCI_AICTRL0) = USEX(SCI_AICTRL1);
			USEX(SCI_AICTRL3) |= CTRL3_BY_NAME;
			OpenFile(0x7fff); /*OpenFile updates AICTRL0 if it finds file*/
			//USEX(SCI_AICTRL3) &= ~CTRL3_BY_NAME; /*cleared by OpenFile() */
			goto checkvalidfile;
#ifdef USE_INFO
		    case 'i':
#endif
		    case 'p': /* 'p'lay by number */
			USEX(SCI_AICTRL0) = atou((void *)&buffer[1]);
		    checkvalidfile:
			if (USEX(SCI_AICTRL0) >= USEX(SCI_AICTRL1)) {
#ifdef USE_PRINTABLE_OUTPUT
			    putstrp("\pnot found ");
			    puthex(USEX(SCI_AICTRL0));
			    putch(' ');
			    putstrpy(fileName);	/*the attribute byte was cleared by OpenFile().*/
			    putch('\n');
#else
			    putstrp("\pnof\n");
#endif
			    break;
			}
#ifdef USE_INFO
			if (buffer[0] == 'i') {
			  OpenFile(USEX(SCI_AICTRL0));
#ifdef USE_PRINTABLE_OUTPUT
			    putdec(USEX(SCI_AICTRL0));
			    putch(' ');
			    putstrpy(fileName);	/*the attribute byte was cleared by OpenFile().*/
			    putch(' ');
			    putdec(fatInfo.fileSize);
			    putch('\n');
#else
			    putword(USEX(SCI_AICTRL0));
			    putch(' ');
			    putstrpy(fileName);	/*the attribute byte was cleared by OpenFile().*/
			    putch(' ');
			    putword(fatInfo.fileSize>>16);
			    putword(fatInfo.fileSize);
			    putch('\n');
#endif
			  break;
			}
#endif
			goto playloop;
		    }
		    bf = buffer;
		prompt:
#ifdef USE_PRINTABLE_OUTPUT
		    putword(('>'<<8)|' ');
#else
		    ;
#endif
		} else {
		    *bf++ = cmd;
		}
	    }
	    /* If OpenFile found read errors. */
	    if (mmc.state != mmcOk || mmc.errors) {
		goto cardchanged;
	    }
	}
    }

 playloop:
#endif /*UART_BUFFERED*/

#ifndef NO_RANDOM /*Disable random play*/
    if ((USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY)) {
#ifdef SHUFFLE_PLAY /*shuffle 23 words*/
      USEX(SCI_AICTRL0) = Shuffle(USEX(SCI_AICTRL1), USEX(SCI_AICTRL0));
#else
      /* random play (not shuffle!) */
      USEX(SCI_AICTRL0) = (myrand() % USEX(SCI_AICTRL1));
#endif
    }
#endif

 reopen:
#ifdef UART_OUTPUT
    goto recstart;
#endif
 
#if 1//ndef UART_BUFFERED //9 words
    /*Wrap from last file to first without trying to open a non-existing file.*/
    {
      register u_int16 ai1 = USEX(SCI_AICTRL1);
      register u_int16 ai0 = USEX(SCI_AICTRL0);
      if (ai0 >= ai1) {
	USEX(SCI_AICTRL0) = ai0-ai1;
      }
    }
#endif
    /* Open the file to play. The file number is in AICTRL0. */
    if (OpenFile(USEX(SCI_AICTRL0) &= 0x7fffU) < 0) { //max 32768 files
      //PlayCurrentFile(); //The function is inlined to save some space.
      //void PlayCurrentFile(void)
      {
	stream_rd_pointer = stream_wr_pointer = (void*)stream_buffer;
	stream_rd_odd = 0;

#if defined(SKIP_ID3V2) /* 48 words */
	/* check for ID3v2 */
	/* An ID3v2 tag can be detected with the following pattern:
	   $49 44 33 yy yy xx zz zz zz zz */
	ReadDiskSectorTo(fragments[0].start, stream_buffer);
	if ((GetLong(0) & 0xffffffL) == 0x334449) {
	  /* skip in 16kB resolution, assume tags are smaller than 2MB */
	  /* 0.5kB resolution now */
	  register u_int16 skip = 32 * GetByte(7) + GetByte(8)/4;
	  /* guard against small fragments */
	  if (skip >= fragments[0].size)
	    skip = fragments[0].size-1;
	  fragments[0].start += skip;
	  fragments[0].size  -= skip;
	  fatInfo.fileSize -= (u_int32)skip * 512;
	}
#endif
	/* Set FILE_READY always */
	{
	  register u_int16 ai3 = USEX(SCI_AICTRL3);
	  ai3 = CTRL3_FILE_READY|(ai3 & ~CTRL3_AT_END);
#if defined(PAUSE_AT_POWERON)
	  /*pause at power-on, then use pause after play*/
	  if ((ai3&CTRL3_PLAY_MODE_MASK)==CTRL3_PAUSE_BEFORE_PLAY){
	    ai3 = CTRL3_PAUSE_AFTER_PLAY | CTRL3_PAUSE_ON;
	  }
#else /*PAUSE_AT_POWERON*/
	  if ((ai3 & CTRL3_PLAY_MODE_MASK) == CTRL3_PAUSE_BEFORE_PLAY) {
	    ai3 |= CTRL3_PAUSE_ON;
	  }
#endif /*else PAUSE_AT_POWERON*/
	  USEX(SCI_AICTRL3) = ai3;
	  spiState = spi_file;
	}
	
#ifdef SAVE_POSITION
	/* Save current song number to EEPROM */
	SaveWord(SAVE_OFFSET+28/*AICTRL0*/, USEX(SCI_AICTRL0));
#endif

#ifdef UART_BUFFERED
	putstrp("\pplay ");
#ifdef USE_DEBUG
	puthex(USEX(SCI_AICTRL0));
#else
	putword(USEX(SCI_AICTRL0));
#endif
	putch(' ');
	putstrpy(fileName); /*the attribute byte was cleared by OpenFile().*/
	putch('\n');
	//putch(fatInfo.fileSize>>24);
	//putch(fatInfo.fileSize>>16);
	//putch(fatInfo.fileSize>>8);
	//putch(fatInfo.fileSize);
#endif

	USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX
#ifdef HAS_KEYS
	  | (1<<INT_EN_TIM0)
#endif
	  ;
	/*
	  For VS1063 we rewrote the main loop.
	  UserCheckTag() and UserAudioFormats() can be eliminated here.
	*/
	parametric_x.playSpeed = 1;
	parametric_x.config1 = 0x0010; /*No implicit upsample--VS1063 20110321*/
	parametric_x.resync = 32767; /* enable resyncs */

	DecodeTimeClear();
	USEX(SCI_DECODE_TIME) = 0;
	USEX(SCI_MODE) &= ~(1<<SCIMB_CANCEL);
	csCancel = 0;
	{
	  __mem_y u_int32 newHead = 0;
	  layer123y.fr.header_change = 3;
	  while (1) {
	    if (CheckAudioFormats(newHead) != 0) {
	      /* One of our codecs decoded something */
	      //newHead = 0; /* Decoded something, start again */
	    }
	    /* Skip a byte and try again. */
	    StreamHeadShift(&newHead);
	    if ((USEX(SCI_MODE) & (1<<SCIMB_CANCEL)) | csCancel)
	      break; //bitwise | is ok here.
	    /* Added: Calls user hook once per byte when searching for audio */
	    UserHook();
	  }
	}
	putstrp("\pdone\n");
      }
      /*end of PlayCurrentFile()*/
    } else {
      //putstrp("\p!f");
      USEX(SCI_AICTRL0) = 0; /* file not found, jump to first file */
      goto cardchanged;
    }
    
    /* If returned from decoding, check if recording was requested. */
    if (record.record == 0x1234) {
    recstart:
#ifdef USE_PRINTABLE_OUTPUT
      putstrp("\pInit Record\n");
#else
      putstrp("\pInit Record\n");
#endif
            /* If record mode, re-enable analog power+drivers immediately. */
    #ifdef PREPARE_RECORDING_INDICATOR
	    USEX(GPIO_ODATA) |= PREPARE_RECORDING_INDICATOR;
    #endif
            USEX(SCI_STATUS) &= ~((1<<SCIST_APDOWN1)|(1<<SCIST_APDOWN2));

	    /*
	      FatPrepareForSaving() goes through FAT allocation table and
	      finds the largest contiguous area. It also goes through the
	      root directory and finds a free directory entry and creates
	      a unique filename that is used later when the file entry is
	      created by FsFatMakeFile().

	      Currently it checks only the 8 character name, so the suffix
	      can be later changed according to the encoding format and
	      the name still remains unique.
	     */
//putstrp("\pprepare");
#ifdef UART_OUTPUT
	    mmc.dirLine = 1; /* just a marker that encoding can be started. */
#else
	    /*
	      FatPrepareForSaving() goes through FAT allocation table and
	      finds the largest contiguous area. It also goes through the
	      root directory and finds a free directory entry and creates
	      a unique filename that is used later when the file entry is
	      created by FsFatMakeFile().
	    */
#ifdef USE_FILEMODE
	    if ((s_int16)record.record < 0 && (fileName[0] & 0xff00)) {
	      /* name given from filemode */
	      FatPrepareForSaving(fileName);
	    } else
#endif
	    {
	      /* not filemode or no name given */
	      FatPrepareForSaving(NULL); /* search free space and a valid name */
	    }
#endif/*elseUART_OUTPUT*/

            if (mmc.dirLine >= 0) { /* found free entry */
                spiState = spi_record;

#if ((CLOCKF_VAL&0xe000) < 0xc000)
		/*at least 4.5x required for 44+k stereo recording*/
		clockX = 9;
#ifdef HAS_KEYS
		USEX(TIMER_CONFIG) = clockX;
#endif
		SetRate(8000);
#endif/*if not high enough already*/

#ifndef UART_OUTPUT
                /* Lower monitor volume during recording. */
		playbackVol = USEX(SCI_VOL);
		USEX(SCI_VOL) += 0x1010;
		USEX(SCI_AICTRL0) = RECORD_FS; /*set in standalone.h*/
#else
		USEX(SCI_VOL) = 0x0101;
		USEX(SCI_AICTRL0) = UART_OUTPUT; //<48k to allow rate adjustment
#endif
		USEX(SCI_AICTRL1) = 0;      //USEX(SCI_WRAM);     /* auto gain*/
		USEX(SCI_AICTRL2) = 0x4000; //USEX(SCI_WRAMADDR); /* max gain */
//		USEX(SCI_WRAMADDR) = 0xe000 + 64; //64kbit/sec
//		USEX(SCI_WRAMADDR) = 0x0005; //q 5

		/* GPIO6 is high -> stereo line */
#if defined(LINE_IN_ENABLE) && LINE_IN_ENABLE
		if ((USEX(GPIO_IDATA) & LINE_IN_ENABLE)) {
		    /* mode: mp3, joint stereo

		       Note: 48kHz sample rate with stereo is too much for
		             3.5x clock. Thus we set 4.5x in record mode.
		     */
#ifdef UART_OUTPUT
		    USEX(SCI_WRAMADDR) = 0xe000 + 192; //192kbit/sec
#else
#ifndef NO_ENCODE_MP3
		    USEX(SCI_WRAMADDR) = 0xe000 + 160; //160kbit/sec
#else
		    USEX(SCI_WRAMADDR) = 6; //quality 6
#endif
#endif
		    //USEX(SCI_WRAMADDR) = 0xe000 + 320; //320kbit/sec
		    USEX(SCI_AICTRL3) =
#ifndef NO_ENCODE_MP3
		      AICTRL3_MP3
#else
		      AICTRL3_VORBIS
#endif
		      //AICTRL3_G722ADPCM
		      | AICTRL3_STEREO
#ifdef UART_OUTPUT
		      | AICTRL3_ENABLEUART
#endif
		      ;
		    USEX(SCI_AICTRL1) = 0x0400;     /* fixed 1.0x gain */
		    USEX(SCI_MODE) |= (1<<SCIMB_LINE); /* Use line input */

#ifdef UART_OUTPUT
		    /* 480600bps @ 4.5x clock */
		    parametric_x.i.encoding.txUartDiv = 0x1306;
		    parametric_x.i.encoding.txUartByteSpeed = 0;
		    parametric_x.i.encoding.txPauseGpio = 0;
#endif
		} else
#endif
		{
		    /* mode: mp3, mono left channel(mic) */
#ifndef NO_ENCODE_MP3
		    USEX(SCI_WRAMADDR) = 0xe000 + 128; //128kbit/sec
#else
		    USEX(SCI_WRAMADDR) = 5; //quality 5
#endif
		    USEX(SCI_AICTRL3) =
#ifndef NO_ENCODE_MP3
		      AICTRL3_MP3
#else
		      AICTRL3_VORBIS
#endif
		      //AICTRL3_G722ADPCM
			//AICTRL3_G711_ALAW
			//AICTRL3_G711_ULAW
			//AICTRL3_PCM
			//AICTRL3_ADPCM
#if 0
			| AICTRL3_MONOMIX /* 4: mono downmix */
#else
			| AICTRL3_LEFT /* 2: left channel, 3: right channel */
#endif
		    | CTRL3_RECORD_ON
		    ;
		    USEX(SCI_MODE) &= ~(1<<SCIMB_LINE); /* Use mic input */
		}
		USEX(SCI_MODE) &= ~(1<<SCIMB_CANCEL);

//putstrp("\precorder");
	        memset(&record, 0, sizeof(record));
		//record.rd = record.wr = record.fill = 0;
		record.record = 0x2468;
		/*
		  NOTE: Encoders() sets applAddr to NULL, but on the other
		        hand it does not get called anyway because encoders
			do not use AudioOutputSamples().
		*/
//putch('e');

#ifdef RECORDING_INDICATOR
                USEX(GPIO_ODATA) |= RECORDING_INDICATOR;
#endif
#ifdef PREPARE_RECORDING_INDICATOR
		USEX(GPIO_ODATA) &= ~PREPARE_RECORDING_INDICATOR;
#endif
		#if 0
		putword(USEX(SCI_MODE));
		putword(USEX(SCI_STATUS));
		putword(USEX(SCI_CLOCKF));
		putword(USEX(SCI_AUDATA));
		putword(USEX(SCI_WRAM));
		putword(USEX(SCI_WRAMADDR));
		putword(USEX(SCI_HDAT0));
		putword(USEX(SCI_HDAT1));
		putword(USEX(SCI_AICTRL0));
		putword(USEX(SCI_AICTRL1));
		putword(USEX(SCI_AICTRL2));
		putword(USEX(SCI_AICTRL3));
		#endif
		USEX(SCI_HDAT0) = USEX(SCI_HDAT1) = 0; //???
		//USEX(INT_ENABLE) = 0; //???
		USEX(SER_DREQ) = 0;
		Disable();
		//InitHardware(); /*no longer clears memory in vs1053/63 so preserves const*/
		//audioPtr.curFctl = audioPtr.newFctl = 0x15555; //correct for 8kHz 20111117
		//memset(&parametric_x, 0, sizeof(parametric_x));
		putstrp("\pCallEncoders\n");
		//putword((u_int16)&sp);
		//memset((void *)0x237d, 0, 0x4000-0x237d);//encoderMp3 etc.
		memset((void *)0x0800, 0, 0x683); //encMAna
#ifndef NO_ENCODE_MP3
		/*The patch is only required for mp3.
		  If you don't use mp3, you can save memory by dropping it. */
                CallEncoders(); /* With mp3 patches from vs1063a-patch.s .
				   loop there until reset */
#else
                CallIROM4(Encoders); /* loop there until reset */
#endif
    }
        }
    goto fileloop;    
}


#if 0
/*Example of a print file function.*/
void printfile() { /* Must have been opened with OpenFile() */
    spix.size = (fatInfo.fileSize+1)/2; /* convert to words */
    spix.fragment = &fragments[0];
    while (1) {
	spix.sector   = spix.fragment->start /*& 0x7fffffffUL*/;
	spix.fragSize = spix.fragment->size;
	do {
	    s_int16 sz = 512;
	    int i;
	    register u_int16 *p;
	    if (sz > spix.size) {
		sz = spix.size;
	    }
	    ReadDiskSectorTo(spix.sector++, stream_buffer);
	    p = stream_buffer;
	    for (i=0;i<sz;i+=2) {
		putword(*p++);
	    }
	    if (sz & 1) {
		putch(*p>>8);
	    }
	    spix.sz -= sz;
	} while (spix.size != 0 && --spix.fragSize > 0);
	if (spix.size == 0 || (s_int32)spix.fragment->start < 0)
	    break;
	spix.fragment++;
    }
}
#endif
