/// \file standalone.c -- VS1063 Stand-Alone MP3 Player Source Code
/// Copyright (C) 2011 VLSI Solution Oy Tampere Finland

/*
  Note:

  DREQ is changed in normal firmware both in SDI interrupts when the
  stream buffer becomes too full, and whenever the stream buffer becomes
  empty enough.

  So, we can use DREQ ourselves as we wish and access to it does not need to
  be protected by interrupt Disable()/Enable(). However, the normal decoding
  will still keep raising DREQ when it sees that there is space for new data.

- VS1063:
  + I2S
  + higher reference
- IN VS1063:
  + SCLK = GPIDATA8
  + XCS  = GPIDATA9
  + SI   = GPIDATA10
  + XDCS = GPIDATA11


- Exiting recoding now from any button. Key scan is not well performed
  during encoding, because applAddr is not called and we have no time base.
  We should use a timer to get 16Hz trigger for key scan.
- Now ends recording nicely and creates 100% valid files.
- If format is WAV, fix the WAV header.
- Finding free space takes a few seconds, so we now cache the free space.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <h1063/encoders.h>
#include "standalone.h"
int csCancel = 0;

/*
  USE_READ_MULTIPLE_BLOCK_ONE is a workaround for
  1. some cards that have issues with implementing READ_SINGLE_BLOCK
     in SPI mode correctly.
  2. cards that have issues implementing READ_MULTIPLE_BLOCK in SPI mode
     correctly.

  Use READ_MULTIPLE_BLOCK, but only read one block at a time.

  See also: asm-readdisk.s
 */
#define USE_READ_MULTIPLE_BLOCK_ONE /* Workaround for some cards that fail with single read block. Is the same size, so unless you use old MMC cards that do not properly support multiple block read, you can keep this defined. If you comment this out, also change #if 0 in asm-readdisk.s to #if 1 */


/* See standalone.h for feature #defines. */

#if defined(SAVE_VOLUME) || defined(SAVE_POSITION)
  #define SAVE_STUFF
#endif
/*
  If any of the key control alternatives, enable other key-related stuff.
 */
#if defined(SCI_KEYS) || defined(COMPAT_KEYS) || defined(DIRECT_KEYS)
  #define LONG_PRESS_TIME 24
  #define SW1 0xffffU
  #define SW2 0x8000U
  #define SW3 0x7fffU

  #define HAS_KEYS
#endif

extern volatile s_int16 *srcWp, *srcRp;

extern __y u_int16 g_dcthi[2048];
auto void play_frame(void);
//void putch(register __a0 int ch);
void putword(register __a0 int ch);

  #define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
  #define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))

#ifndef NOCHECKFILETYPE
  #define CHECKFILETYPE CheckFileType1063
#endif/*NOCHECKFILETYPE*/


s_int16 hcShift; //asm-readdisk uses hcShift

#ifndef GPIO_IDATA
  #define GPIO_IDATA GPIO_DATA
#endif


#define SPI_xCS   1 /* GPIO */
#define SPI_CLK   8
#define SPI_MISO  4
#define SPI_MISO_SHIFT -3
#define SPI_xCS2  2
#define SPI_MOSI  1 /* DREQ */

#ifdef UART_BUFFERED
static int fileMode =
#ifdef START_IN_FILEMODE
  1
#else
  0
#endif
  ;
#endif

#ifdef USE_DEBUG
static char hex[] = "0123456789ABCDEF";
void puthex(u_int16 d) {
    register i;
    for (i=0;i<4;i++) {
	putch(hex[(d>>12)&15]);
	d <<= 4;
    }
}
void puthex2(u_int16 d) {
    register i;
    putch(hex[(d>>4)&15]);
    putch(hex[(d>>0)&15]);
}
#endif


typedef enum {
    spi_pause=0,
    spi_file=1,
    spi_seek=2,
    spi_waitdata=3,
    spi_data=4,
    spi_zeros=5,
    spi_record=6,
    spi_waitima=7
} SPISTATE;

SPISTATE spiState, oldSpiState;
u_int16 spiCnt;


struct SPIX { /* do not touch! ASM uses the structure */
    s_int32 size;
    u_int32 sector;
    u_int16 fragSize; /* TODO: make this 32-bit and multiply fragment list clusters by sectorsPerCluster. */
    __mem_y struct FRAGMENT *fragment;
} spix;


extern __mem_y u_int16 hwSampleRate;

/* Can be done for all formats, because IdleHook is run as the main thread
   and not in interrupts. Whenever IdleHook is called, there either is
   not enough stream data anyway, or the decoder is outputting samples
   and there is not enough space in the audio buffer (thus not reading
   stream). */
u_int16 __x * __x stream_save;

#ifdef HAS_KEYS
__mem_y struct {
    u_int16 Data; /* must be first entry! Used by c.s */
    u_int16 Cnt;
    u_int16 Strobe;
    u_int16 Timer; /* Data,Cnt,Strobe,Timer must be in this order (buts.s) */
    u_int16 Key;
    s_int16 KeyTime;
} ui;
#endif/*HAS_KEYS*/



void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
#ifdef SAVE_VOLUME
int volChanged = 0;
#endif

#ifdef SAVE_STUFF
  #define SPI_WREN  0x06
  #define SPI_WRDI  0x04
  #define SPI_RDSR  0x05
  #define SPI_WRSR  0x01
  #define SPI_READ  0x03
  #define SPI_WRITE 0x02
  
  register __a1 short SpiDelay(void);
  auto u_int16 SpiSendReceiveSpi(register __a0 u_int16 datTopAligned, register __a1 s_int16 bits);
  
  #if 1
  void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
  #else
  void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value) {
      USEX(GPIO_ODATA) &= ~(SPI_xCS2|SPI_xCS);
      SpiSendReceiveSpi(SPI_WREN<<8, 8);
      USEX(GPIO_ODATA) |= SPI_xCS;
      USEX(GPIO_ODATA) &= ~SPI_xCS;
      SpiSendReceiveSpi(SPI_WRITE<<8, 8);
      SpiSendReceiveSpi(addr/*volume=26*/, 16);
      SpiSendReceiveSpi(value, 16);
      USEX(GPIO_ODATA) |= SPI_xCS|SPI_xCS2;//SPI_CLK;
  }
  #endif
  //SaveWord(26/*volume=26*/, USEX(SCI_VOL));
#endif

#ifdef SHUFFLE_PLAY
u_int16 shuffle_offset = 0;
#endif


/*
  Called before HALT
 */
void UserHook(void) {
#if 1
    /* Sometimes when audio buffer is empty when samplerate gets changed,
       SetRate() in vs1063a overwrites interrupt's change of the high 4
       bits of samplerate control value. */
    if ((USEX(FREQCTLH) & 15) != (audioPtr.curFctl >> 16)) {
	Disable();
	USEX(FREQCTLH) = (USEX(FREQCTLH) & ~15) | (audioPtr.curFctl >> 16);
	Enable();
    }
#endif

    #define SW1 0xffffU
    #define SW2 0x8000U
    #define SW3 0x7fffU
    //#define SW4 0x4000U
      
#ifdef HAS_KEYS
      /* -- KEY USER INTERFACE -- */
      if (ui.Strobe) {
	ui.Data = ui.Strobe = 0;
#ifdef SCI_KEYS /*If both keys and SCI control.*/
	/* SCI-controlled version, but with additional key support.
	   GPIO4 for Pause/Play(short press) and Toggle Shuffle(long press), 
	   GPIO5 for Next(short) and Vol+(long), 
	   GPIO6 for Previous(short) and Vol-(long). */
#if 1
	ui.Data = USEX(GPIO_IDATA) & (7<<4);/*low is idle, high when pressed*/
#else
	ui.Data = ~USEX(GPIO_IDATA) & (7<<4); /*high idle, low when pressed*/
#endif
	if (ui.Data == (1<<6)) {
	    ui.Data = SW2;/*prev/vol-*/
	} else if (ui.Data == (1<<5)) {
	    ui.Data = SW1;/*next/vol+*/
	} else if (ui.Data == (1<<4)) {
	    ui.Data = SW3;/*pause/play/loudness*/
	}
#endif/*SCI_KEYS*/

#ifdef COMPAT_KEYS
	/*
	  Read the 3 buttons on the VS10xx prototyping board.
	*/
	if (!(USEX(GPIO_IDATA) & (1<<10)/*SI*/)) {
	    ui.Data = SW2;/*prev/vol-*/
	} else if (!(USEX(GPIO_IDATA) & (1<<11)/*XDCS*/)) {
	    ui.Data = SW1;/*next/vol+*/
	} else {
	    register u_int16 t = USEX(GPIO_ODATA);
	    USEX(GPIO_ODATA) = t & ~SPI_xCS;
	    /* Should have a short wait here, but seems to work this way.*/
	    if (!(USEX(GPIO_IDATA) & (1<<11)/*XDCS*/)) {
		ui.Data = SW3;/*play/pause/ etc*/
	    } else if (!(USEX(GPIO_IDATA) & (1<<9)/*XCS*/)) {
		ui.Data = SW3;
	    }
	    USEX(GPIO_ODATA) = t;
	}
#endif/*COMPAT_KEYS*/
	
#ifdef DIRECT_KEYS
	/*
	  Read keys directly connected to SCI pins.
	  Idle high, button pulls low.
	  + SCLK = GPIDATA8 (unused)
	  + XCS  = GPIDATA9
	  + SI   = GPIDATA10
	  + XDCS = GPIDATA11
	*/
	if (!(USEX(GPIO_IDATA) & (1<<10)/*SI*/)) {
	    ui.Data = SW2;/*next*/
	} else if (!(USEX(GPIO_IDATA) & (1<<11)/*XDCS*/)) {
	    ui.Data = SW1;/*prev*/
	} else if (!(USEX(GPIO_IDATA) & (1<<9)/*XCS*/)) {
	    ui.Data = SW3;/*pause/play*/
	}
	//else if (!(USEX(GPIO_IDATA) & (1<<8)/*SCLK*/)) {
	//ui.Data = SW1;/*prev*/
	//}
#endif/*DIRECT_KEYS*/

	// -- KEY FUNCTIONALITY --
	if (ui.Data == ui.Key) { // New Keystate is same as previous Keystate
	  if (ui.KeyTime < 12) {
	    ui.KeyTime++;
	  } else // PKP: keystate has been consistent for 12 ticks - it's a long keypress
	    if (ui.Key && ui.KeyTime > 0) {
	      /* long presses */
	      register u_int16 vol = USEX(SCI_VOL);
	      if (ui.Key == SW1) {
#ifdef UART_BUFFERED
		goto volup;
#else
		if (vol) {
		  vol -= 0x0101;
		  goto setvol;
		}
#endif/*elseUART_BUFFERED*/
	      } else if (ui.Key == SW2) {
#ifdef UART_BUFFERED
		goto voldown;
#else
		if (vol != (s_int16)0xffffU)
		  vol += 0x0101;
	      setvol:
		USEX(SCI_VOL) = vol; /*in VS1011E disables ints! */
#ifdef SAVE_VOLUME
		volChanged = 1;
#endif/*SAVE_VOLUME*/
#endif/*elseUART_BUFFERED*/
	      } else {
		/* SW3 long press */
		if (ui.KeyTime == 8) {
		  ui.KeyTime = 9;
#if defined(LOUDNESS_ALWAYS_ON_OFF)
		  /* if pause on -- toggle random play */
		  USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
		  if (spiState == spi_pause) {
		    spiState = oldSpiState;
		    stream_wr_pointer = stream_save;
		  }
#else /*LOUDNESS_ALWAYS_ON_OFF*/
		  
		  if (spiState
#ifdef SW3_RANDOM_PLAY
		      !=
#else
		      ==
#endif/*SW3_RANDOM_PLAY*/
		      spi_pause) {
		    /* if pause on -- toggle random play */
		    USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
		  } else {
		    /* toggle loudness - have LSb set when ON */
		    USEX(SCI_BASS) ^= USEX(SCI_AICTRL2);
		  }
#endif/*else LOUDNESS_ALWAYS_ON_OFF*/
		}
	      }
	    }
	} else { // PKP: Keystate is not the same as previous keystate
	  // PKP: Handle short keypresses
	  if (ui.Data == 0 && ui.KeyTime > 0) {
	    // PKP: Handle end of short keypress (current keystate ui.Data=0)
	    /* short presses */
	    if (ui.KeyTime < 8) {
	      /* was pressed < 0.5 seconds */
	      
	      if (ui.Key == SW3) {
		/* play pause/continue -- not very good for midi.. */
		USEX(SCI_AICTRL3) ^= CTRL3_PAUSE_ON;
	      } else if (ui.Key == SW1 || ui.Key == SW2) {
		if (ui.Key == SW1) {
		  ui.Key = ui.Data;
		  #ifdef UART_BUFFERED
		  goto nextsong;
                  #else
		  USEX(SCI_AICTRL0) += 0x8001;
                  #endif/*elseUART_BUFFERED*/
		} else {
		  ui.Key = ui.Data;
		  #ifdef UART_BUFFERED
		  goto prevsong;
		  #else
		  if (USEX(SCI_DECODE_TIME) < 4) {
		      if (USEX(SCI_AICTRL0) == 0) {
			  USEX(SCI_AICTRL0) = USEX(SCI_AICTRL1) + (0x8000-1);
		      } else {
			  USEX(SCI_AICTRL0) += (0x8000-1);
		      }
		  } else {
		      USEX(SCI_AICTRL0) |= 0x8000U;
		  }
		  #endif/*elseUART_BUFFERED*/
		}
	      }
	    }//(ui.keyTime was < 12)
	  }
	  ui.Key = ui.Data;
	  ui.KeyTime = 0;
	}
        
#if defined(SAVE_VOLUME)
	/*Save volume when volume button is released. */
	if (ui.Data == 0 && volChanged && spiState == spi_seek) {
	  volChanged = 0;
	  SaveWord(26/*volume=26*/, USEX(SCI_VOL));
	}
#endif/*SAVE_VOLUME*/
      }
    #endif /*HAS_KEYS*/
    
#ifdef UART_BUFFERED /* This is becoming the new default. */
    if (UartFill() != 0) {
      int c = UartGetByte();
      if (c == 0xef) {
	USEX(INT_ENABLE) = 0;
	Monitor();
      } else if (c == 'f') { /* filemode */
	fileMode = 1;
      } else if (c == 'c') { /* continuous play mode - random off */
#ifndef NO_RANDOM
	USEX(SCI_AICTRL3) &= ~CTRL3_RANDOM_PLAY;
#endif
	fileMode = 0;
      } else if (c == ',') { /* previous */
      prevsong:
	if (USEX(SCI_DECODE_TIME) < 4) {
	  if (USEX(SCI_AICTRL0) == 0) {
	    USEX(SCI_AICTRL0) = USEX(SCI_AICTRL1) + (0x8000-1);
	  } else {
	    USEX(SCI_AICTRL0) += (0x8000-1);
	  }
	} else {
	  USEX(SCI_AICTRL0) |= 0x8000U; //restart the same file
	}
      } else if (c == 'C') { /* Cancel */
	USEX(SCI_AICTRL0) |= 0x8000;
#ifndef NO_RANDOM
      } else if (c == 'r') { /* random play */
	USEX(SCI_AICTRL3) |= CTRL3_RANDOM_PLAY;
#endif
#ifdef USE_EARSPEAKER
      } else if (c == 'e') {
	parametric_x.earSpeakerLevel ^= 0x6000;
#endif
      } else if (c == 'l') {
	if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) == CTRL3_LOOP_SONG) {
	  USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
	  putch('>');
	} else {
	  USEX(SCI_AICTRL3) =
	    (USEX(SCI_AICTRL3) & ~CTRL3_PLAY_MODE_MASK) | CTRL3_LOOP_SONG;
	  putch('l');
	}
      } else if (c == '.') { /* next */
      nextsong:
	USEX(SCI_AICTRL0) += 0x8001;
      } else if (c == '=') { /* pause */
#ifdef USE_PRINTABLE_OUTPUT
	putstrp("\ppause\n");
#else
	putch('=');
#endif
	USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
      } else if (c == '>') { /* play 1x */
	USEX(SCI_AICTRL3) &= ~CTRL3_PAUSE_ON;
	parametric_x.playSpeed = 1;
#ifdef USE_PRINTABLE_OUTPUT
	putstrp("\pplay\n");
#else
	putch('>');
#endif
      } else if (c == 0xbb /* '»' */) {
	/* play faster */
	parametric_x.playSpeed++;
#ifdef USE_PRINTABLE_OUTPUT
	putdec(parametric_x.playSpeed);
	putch('\n');
#else
	putch(parametric_x.playSpeed);
#endif
      } else if (c == '+') { /* volume up */
      volup:
	if (USEX(SCI_VOL))
	  USEX(SCI_VOL) -= 0x0101;
	goto setvol;
      } else if (c == '-') { /* volume down */
      voldown:
	if (USEX(SCI_VOL) != 0xffffU)
	  USEX(SCI_VOL) += 0x0101;
      setvol:
#ifdef SAVE_VOLUME
	volChanged = 1;
#endif/*SAVE_VOLUME*/
#ifdef USE_PRINTABLE_OUTPUT
	putdec(USEX(SCI_VOL) & 255);
	putch('\n');
#else
	putch(USEX(SCI_VOL));
#endif
      } else if (c == '?') { /* decode time + info -- Note: will block.*/
#ifdef USE_PRINTABLE_OUTPUT
	putdec(USEX(SCI_DECODE_TIME));
	putch(' ');
	/* play file position from 0..255 */
	putdec((fatInfo.fileSize-2*spix.size) / ((fatInfo.fileSize >> 8)+1));
	putch('\n');
#else
	putword(USEX(SCI_DECODE_TIME));
	/* play file position from 0..255 */
	putch((fatInfo.fileSize-2*spix.size) / ((fatInfo.fileSize >> 8)+1));
#endif
      } else {
	/* unknown command */
	//putch(c);
      }
    }
#endif/*UART_BUFFERED*/

    
      /* play pause/continue */
      if ((USEX(SCI_AICTRL3) & CTRL3_PAUSE_ON)) {
	/* CTRL3_PAUSE_ON overlaps the record pause bit. */
	parametric_x.playMode |= PLAYMODE_PAUSE_ON;
      } else {
	parametric_x.playMode &= ~PLAYMODE_PAUSE_ON;
      }
      if ((USEX(SCI_AICTRL0) & 0x8000U)) {
        /* new song */
        USEX(SCI_AICTRL0) &= 0x7fff; /* ack command */
    out_of_wav:
	USEX(SCI_MODE) |= (1<<SCIMB_CANCEL);
	csCancel = 1;
	parametric_x.resync = 0; /* disable resyncs - csCancel works better. */

#ifdef USE_READ_MULTIPLE_BLOCK_ONE
	MmcStopTransmission();
	SpiSendClocks();
#else
	/* Discard any unread data, including CRC, before starting again */
	{
	  register int i;
	  for (i=0; i<256+2; i++)
	    SpiSendReceiveMmc(-1/*0xffff*/,16);
	}
	SpiSendClocks();
#endif
	/* Clear pause and file ready bits automatically */
	USEX(SCI_AICTRL3) &= ~(CTRL3_PAUSE_ON|CTRL3_FILE_READY|CTRL3_AT_END);
	spiState = spi_zeros; /* continue with play cancel procedure */
	spiCnt = 0;
      }
    
    
//#define PLAYTIME 15 /*play only this number of seconds (15) of each file*/

    #ifdef PLAYTIME
      if (USEX(SCI_DECODE_TIME) > PLAYTIME && spiState != spi_zeros) {
          USEX(SCI_AICTRL0)++;
	  USEX(SCI_MODE) |= (1<<SCIMB_CANCEL);
          spiState = spi_zeros;
      }
    #endif/*PLAYTIME*/
    
    if (spiState == spi_file) { /* Play file from the beginning. */
        /* extra byte will be read if the file size is odd */
        spix.size = (fatInfo.fileSize+1)/2; /* convert to words */
        spiState = spi_seek;
	//setfrag:
        spix.fragment = &fragments[0];
        spix.sector   = fragments[0].start /*& 0x7fffffffUL*/;
        spix.fragSize = fragments[0].size; /*todo: *fatInfo.sectorsPerCluster*/
    }
    //checkseek:
    if (spiState == spi_seek) {
#ifdef USE_READ_MULTIPLE_BLOCK_ONE
	MmcCommand(MMC_READ_MULTIPLE_BLOCK|0x40,
	  (spix.sector & 0x7fffffffUL) << mmc.hcShift);
#else
	/* Send the SD/MMC read block command */
	MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,
		   (spix.sector & 0x7fffffffUL) << mmc.hcShift);
#endif
        spiState = spi_waitdata;
        spiCnt = 0;
    }

    /*
      We have sent the read block command and now check if the data
      is available to be read out.
     */
    if (spiState == spi_waitdata) {
	register u_int16 res = SpiSendReceiveMmc(0xff00,8);
	if (res == 0xfe) { /* 0xfe is the data start token */
	    spiState = spi_data; /* go to fetch mode */
	    spiCnt = 0;
	    NextSector();
	} else if (res == 0xff) { /* card is still fetching block */
	    if ((u_int16)++spiCnt > 50000) {
		MyReset(); /* timeout, restart immediately */
	    }
	} else {
	    /* Otherwise we have Read Error, and data is not transferred */
	  csCancel = 1;
	    goto end_of_file;
	}
    } else if (spiState == spi_data) { /* data transfer state */
	/* Loop until the whole file is read, or stream buffer is full,
	   or the next sector needs to be read. */
	while (spix.size > 0) {
	    register int i, j = 32;
	    register u_int16 *wr;
	    /* FLAC has larger stream buffer */
	    if (USEX(SCI_HDAT1) == (('f'<<8) | 'L')) {
		register s_int16 t = stream_wr_pointer - stream_rd_pointer;
		if (t < 0)
		    t += 0x1800;
		if (t > 0x1800/*STREAM_BUFFER_SZ*/-40)
		    break; /* stream buffer full, leave loop */
		wr = (void *)stream_wr_pointer;
		if (j > spix.size)
		    j = spix.size;
		for (i=0;i<j;i++) {
		    *wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
		    if (wr >= &stream_buffer[0x1800/*STREAM_BUFFER_SZ*/])
			wr = stream_buffer;
		}
	    } else {
		if (StreamDiff() >= STREAM_BUFFER_SZ-40)
		    break; /* stream buffer full, leave loop */
		wr = (void *)stream_wr_pointer;
		if (j > spix.size)
		    j = spix.size;
		for (i=0;i<j;i++) {
		    *wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
		    if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
			wr = stream_buffer;
		}
	    }
	    stream_wr_pointer = (void *)wr;
	    spix.size -= j;//32;
	    spiCnt += 32;
	    if (spiCnt >= 256) {
		/* 256 words read, get next block in file */
		spiState = spi_seek;
		SpiSendReceiveMmc(-1/*0xffff*/,16); /* discard crc */
#ifdef USE_READ_MULTIPLE_BLOCK_ONE
		MmcStopTransmission();
#endif
#if 0
		if (USEX(SCI_HDAT1) == (('f'<<8) | 'L')) {
		    goto checkseek;
		}
#endif
		break;
	    }
	    /* watchdog clearing here doesn't work with midi */
	}
  
        if (spix.size <= 0) {
end_of_file:
#ifdef USE_READ_MULTIPLE_BLOCK_ONE
	  MmcStopTransmission(); /* Stop Transmission */
	  SpiSendClocks();
#else
	  /* Discard any unread data, including CRC, before starting again */
	  {
	    register int i;
	    for (i=0; i<256+2; i++)
	      SpiSendReceiveMmc(-1/*0xffff*/,16);
	  }
	  SpiSendClocks();
#endif

	  {
	    register int m = (USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK);
	    if (m == CTRL3_LOOP_SONG){
	      /* restart the same file immediately */
	      spiState = spi_file;
	    } else {
	      /*NEW: pause after play */
	      if (m == CTRL3_PAUSE_AFTER_PLAY) {
		USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON | CTRL3_AT_END;
	      } else {
		USEX(SCI_AICTRL0) += 1; /* next file */
	      }
	      spiState = spi_zeros;
	      //putch('Z');
	      spiCnt = 0;
	      if (USEX(SCI_HDAT1) == (('f'<<8) | 'L'))
		USEX(SCI_MODE) |= (1<<SCIMB_CANCEL);
	    }
	  }
        }
        #ifdef NO_WMA /* reject WMA files */
            if (USEX(SCI_HDAT1) == (('W'<<8)|'M')) {
		USEX(SCI_AICTRL0) += 1; /* next file */
		goto goreset;//MyReset();
	    }
        #endif
        #ifdef NO_AAC /* reject AAC files */
	    {
		register u_int16 t = USEX(SCI_HDAT1);
		//0x4154(AT) 0x4144(AD) 0x4d34(M4)
		if (t == 0x4154 || t == 0x4144 || t == 0x4d34) {
		    USEX(SCI_AICTRL0) += 1; /* next file */
		    goto goreset;//MyReset();
		}
	    }
        #endif
    } else if (spiState == spi_zeros) { /* send zeros after the file */
#ifdef USE_READ_MULTIPLE_BLOCK_ONE
      //
#else
	if (spiCnt == 0) {
            /* Discard any unread data, including CRC, before starting again */
	    register int i;
	    for (i=0; i<256+2; i++)
		SpiSendReceiveMmc(-1/*0xffff*/,16);
	    SpiSendClocks();
	}
#endif
        /* Workaround for WMA resync after stream ended! */
        parametric_x.resync = 0;
	/* FLAC files use a bigger stream buffer,
	   also requires 12k of endFillByte! */
	if (USEX(SCI_HDAT1) == (('f'<<8) | 'L')) {
	    register s_int16 t = stream_wr_pointer - stream_rd_pointer;
	    if (t < 0)
		t += 0x1800;
	    if (t < 0x1800/*STREAM_BUFFER_SZ*/-40) {
		register int i, z = parametric_x.endFillByte * 0x0101;
		register u_int16 *wr = (void *)stream_wr_pointer;
		for (i=0;i<32;i++) {
		    *wr++ = z;
		    if (wr >= &stream_buffer[0x1800/*STREAM_BUFFER_SZ*/])
			wr = stream_buffer;
		    ++spiCnt;
		    stream_wr_pointer = (void *)wr;
		}
	    }
	} else {
	    if (StreamDiff() < STREAM_BUFFER_SZ-40) {
		register int i, z = parametric_x.endFillByte * 0x0101;
		register u_int16 *wr = (void *)stream_wr_pointer;
		for (i=0;i<32;i++) {
		    *wr++ = z;
		    if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
			wr = stream_buffer;
		    ++spiCnt;
		    stream_wr_pointer = (void *)wr;
		}
	    }
	}
	if (spiCnt >= 256+1*1024) {
	  /* Cancel play at the end of the song. */
	  USEX(SCI_MODE) |= (1<<SCIMB_CANCEL);
	  csCancel = 1;
	}
	if (USEX(SCI_HDAT1) == (('f'<<8) | 'L') && spiCnt >= 256+6*1024) {
	    goto goreset;
	} else if (/*USEX(SCI_HDAT1) == 0 ||
		   (USEX(SCI_HDAT1) == 0x7665 && spiCnt >= 256+2*64)
		   ||*/ spiCnt >= 256+4*2048) {
	    /* 2050 bytes */
goreset:
	  /*in effect we always end up here!*/
	  putstrp("\pcancel too slow!\n");
//putch('e');
	    MyReset();
	}
    } else {
//putch('?');
    }
    

#ifdef HAS_KEYS
#ifdef NO_DREQ_LED
    /* Save space by not updating DREQ. */
#else /*NO_DREQ_LED*/
    /* 15 words */
    {
      register int t = USEX(SCI_AICTRL3) & 1;  /* lowest bit is global random play*/
      /* DREQ indicates random play status */
      /*
	Play:       XXRR............ 0x1c!=0 -> 0 0x1c==0->1
	Play Pause: XR......XR...... 0x0e!=0 -> 0 0x0e==0->1
	RecPause    XXXX....XXXX.... 0x0e!=0 -> 0, otherwise 1
	Record      XXXX............ 0x1c!=0 -> 0, otherwise 1
      */
      u_int16 mask = 0x1f;
      if ((USEX(SCI_AICTRL3) & AICTRL3_PAUSEON)
	  || spiState == spi_pause) {
	mask = 0x3f;
      }
      mask = ui.Timer & mask; /*ui.Timer only available with keys.*/
      if (mask <= 4) {
	t = 1;
      } else if (/*spiState == spi_waitima ||*/ mask >= 8) {
	t = 0;
      }
      USEX(SER_DREQ) = t;
    }
#endif/*!NO_DREQ_LED*/
#endif/*HAS_KEYS*/
}



void NewSinTest(void);

u_int16 sectorBuffer[512/2] = {0};


void PlayCurrentFile(void) {
    stream_rd_pointer = stream_wr_pointer = (void*)stream_buffer;
    stream_rd_odd = 0;

#if defined(SKIP_ID3V2) /* 48 words */
    /* check for ID3v2 */
    /* An ID3v2 tag can be detected with the following pattern:
       $49 44 33 yy yy xx zz zz zz zz */
    ReadDiskSectorTo(fragments[0].start, stream_buffer);
    if ((GetLong(0) & 0xffffffL) == 0x334449) {
      /* skip in 16kB resolution, assume tags are smaller than 2MB */
      /* 0.5kB resolution now */
      register u_int16 skip = 32 * GetByte(7) + GetByte(8)/4;
      /* guard against small fragments */
      if (skip >= fragments[0].size)
	skip = fragments[0].size-1;
      fragments[0].start += skip;
      fragments[0].size  -= skip;
      fatInfo.fileSize -= (u_int32)skip * 512;
    }
#endif

    /* Set FILE_READY always */
    USEX(SCI_AICTRL3) = CTRL3_FILE_READY | (USEX(SCI_AICTRL3) & ~CTRL3_AT_END);
	
#if defined(PAUSE_AT_POWERON)
    /*pause at power-on, then use pause after play*/
    if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) == CTRL3_PAUSE_BEFORE_PLAY){
	USEX(SCI_AICTRL3) = CTRL3_PAUSE_AFTER_PLAY | CTRL3_PAUSE_ON;
    }
#else /*PAUSE_AT_POWERON*/
    if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) == CTRL3_PAUSE_BEFORE_PLAY) {
      USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
    }
#endif /*else PAUSE_AT_POWERON*/

    spiState = spi_file;
	
#ifdef SAVE_POSITION
    /* Save current song number to EEPROM */
    SaveWord(SAVE_OFFSET+28/*AICTRL0*/, USEX(SCI_AICTRL0));
#endif

#ifdef UART_BUFFERED
    parametric_x.playSpeed = 1;
    putstrp("\pplay ");
    #ifdef USE_DEBUG
    puthex(USEX(SCI_AICTRL0));
    #else
    putword(USEX(SCI_AICTRL0));
    #endif
    putch(' ');
    putword(fileName[0]);
    putword(fileName[1]);
    putword(fileName[2]);
    putword(fileName[3]);
    putword(fileName[4]);
    putch(fileName[5]>>8);
    putch('\n');
    //putch(fatInfo.fileSize>>24);
    //putch(fatInfo.fileSize>>16);
    //putch(fatInfo.fileSize>>8);
    //putch(fatInfo.fileSize);
#endif

    USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX
#ifdef HAS_KEYS
      | (1<<INT_EN_TIM0)
#endif
      ;
    /*
      For VS1063 we rewrote the main loop.
      UserCheckTag() and UserAudioFormats() can be eliminated here.
    */
    parametric_x.config1 = 0x0010; /*No implicit upsample -- VS1063 20110321*/
    parametric_x.resync = 32767; /* enable resyncs */

    DecodeTimeClear();
    USEX(SCI_DECODE_TIME) = 0;
    USEX(SCI_MODE) &= ~(1<<SCIMB_CANCEL);
    csCancel = 0;
    {
	__mem_y u_int32 newHead = 0;
	layer123y.fr.header_change = 3;
	while (1) {
#if 0
	    newHead = UserCheckTag(newHead);
	    if (UserAudioFormats(newHead) != 0) {
		/* User codec decoded something */
		newHead = 0;
	    } else 
#endif
	    if (CheckAudioFormats(newHead) != 0) {
		/* One of our codecs decoded something */
		newHead = 0; /* Decoded something, start again */
	    }
	    /* Skip a byte and try again. */
	    StreamHeadShift(&newHead);

	    if (csCancel || (USEX(SCI_MODE) & (1<<SCIMB_CANCEL)))
	      break;
	    /* Added: Calls user hook once per byte when searching for audio */
	    UserHook();
	}
    }
    putstrp("\pdone\n");
}



#if 0 //Bring a default version from the library
void IterateFilesCallback(register __b0 u_int16 *name) {
    register int i;
    /* Dump the whole FAT directory entry */
    for (i=0;i<32/2;i++) {
        putword(*name++);
    }
}
#endif

//#define TEST_POWERUP
#ifdef TEST_POWERUP
/*
  TODO: use a timer interrupt for a watchdog timer.
 */
extern s_int16 flacTmpBuf[64];
extern __mem_y u_int16 bassReg, volReg;
void TestPowerUp() {
  USEX(GPIO_DDR) |= (1<<4);
    /* Make sure we have the right upper ROM visible. */
    USEX(VS1053_IROM4) &= ~(IROM4_ENABLE);
    hwSampleRate = 11; /*force, mem is only cleared at boot, not restart*/
//    uartByteSpeed = 960; //for vs1063
    uartByteSpeed = 115200/10; //for vs1063
  USEX(GPIO_ODATA) ^= (1<<4);
    InitHardware(); /*no longer clears memory in vs1053/63 so preserves const*/
    audioPtr.curFctl = audioPtr.newFctl = 0x15555; //correct for 8kHz 20111117
#if 1 /*set 48kHz */
    AudioSet(1, 48000U);
    USEX(FREQCTLH) = (USEX(FREQCTLH) & ~(15|FCH_PLL_LOCKED)) | (u_int16)(audioPtr.newFctl>>16);
    USEX(FREQCTLL) = audioPtr.newFctl;
    audioPtr.curFctl = audioPtr.newFctl;
    audioPtr.changeTimer = -1;
#endif
    
    //2 = drivers, 1 = anapwr
    //USEX(SCI_STATUS) &= ~((1<<SCIST_APDOWN2)|(1<<SCIST_APDOWN1)); // bad
    //USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN2); //drivers first - bad
    // USEX(DAC_VOL) = 0x6666;
    USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1); //powers first - best
    //putstrp("\pstarting");

#if 0
    CheckVolumeAndBass();
    USEX(DAC_VOL) = 0x0c0c; //override volume
    memset(flacTmpBuf, 24000, sizeof(flacTmpBuf));
    flacTmpBuf[7] = flacTmpBuf[6] = flacTmpBuf[3] = flacTmpBuf[2] = -24000;
    {
      register int i;
      for (i=0;i<100;i++) {
	WmaStereoCopy(flacTmpBuf, 4);
      }
    }
    memset(flacTmpBuf, 0, sizeof(flacTmpBuf));
    WmaStereoCopy(flacTmpBuf, 4);
#endif
    USEX(INT_ENABLE) |= (1<<INT_EN_DAC);
  USEX(GPIO_ODATA) ^= (1<<4);
  audioPtr.timeToRemovePDown = 2048;//AUDIO_BUFFER_SZ/2;

    USEX(INT_GLOB_ENA) = 0;
    USEX(INT_GLOB_ENA) = 0;
    USEX(INT_GLOB_ENA) = 0;
    while (1) {
      memset(flacTmpBuf, 0, sizeof(flacTmpBuf));
      WmaStereoCopy(flacTmpBuf, 4);
      if (USEX(SCI_STATUS) & (1<<SCIST_APDOWN2)) {
	USEX(GPIO_ODATA) |= (1<<4);
      } else {
	USEX(GPIO_ODATA) &= ~(1<<4);
	break;
      }
    }
    bassReg ^= 1;
    CheckVolumeAndBass();
    while (1)
      ;
}
#endif

void flacStreamBufferRestore(void);
auto void MyMain(register s_int16 __a1 fullReset) {
    u_int16 tmp;
    applAddr = NULL;
    USEX(SCI_AIADDR) = 0;
#ifdef ENABLE_HIGHREF
    USEX(SCI_STATUS) |= (1<<SCIST_REFERENCE_SEL); /*1.65V*/
#endif
//    USEX(SCI_CLOCKF) = CORE_CLOCK_3_5X;
    /* Install new idle hook handler */
    WriteIMem((u_int16)IdleHook, 0x28000000UL | ((u_int32)(u_int16)UserHook<<6));
#ifdef TEST_POWERUP
    if(fullReset) {
      TestPowerUp();
    }
#endif
      
  #ifdef UART_BUFFERED
    //Not really supported yet in the player!
    UartFill(); /* pull in UART stuff from the library */
  #endif

    /* Make sure we have the right upper ROM visible. */
    USEX(VS1053_IROM4) &= ~(IROM4_ENABLE);
    USEX(SCI_HDAT1) = 0; /* clear so FLAC is not indicated. */
    /* Just in case, restore the normal SDI interrupt handler (1024 words). */
    flacStreamBufferRestore();
    /* clear mp4ASC, seems to be needed after FLAC. */
    memset((void *)0x2000, 0, 64);

    /* no files yet */
    USEX(SCI_AICTRL1) = 0;

    /* For MMC/SD handling the SCI_NEWMODE must be set, and also
       the relevant GPIO pins configured correctly. */

#define LINE_IN_ENABLE (1<<6)
#define RECORDING_INDICATOR (1<<4)
#define PREPARE_RECORDING_INDICATOR (1<<5)
    USEX(SCI_MODE) = (USEX(SCI_MODE) & ~(1<<SCIMB_CANCEL)) |
	(1<<SCIMB_SDI_NEWMODE) | (1<<SCIMB_LINE)
#ifdef ENABLE_LAYER12
	| (1<<SCIMB_ALLOW_LAYER12)
#endif
	;
    USEX(GPIO_ODATA) = SPI_xCS|SPI_CLK|SPI_xCS2|SD_POWER_ENABLE;
    USEX(GPIO_DDR)   = SPI_xCS|SPI_CLK|SPI_xCS2|SD_POWER_ENABLE
    #ifdef RECORDING_INDICATOR
	| RECORDING_INDICATOR
    #endif
    #ifdef PREPARE_RECORDING_INDICATOR
	| PREPARE_RECORDING_INDICATOR
    #endif
    #if defined(ENABLE_I2S)
	/* Enable I2S pins */
	| 0xf0;
      USEX(I2S_CONTROL) = 0x0c
    #endif
	;

    /* InitHardware() configures clock/PLL according to SCI_CLOCKF,
       initializes stream and audio buffer pointers, sets default
       samplerate (8000Hz), clears HDAT0,HDAT1,DECODE_TIME,AUDATA,
       and sets the cosmetic chip ID in parametric_x.chipID .

       User should set uartByteSpeed 960 before the call to get the
       default 9600bps transfer rate. (vs1063)
    */
#if 0
    USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1);
#endif
    hwSampleRate = 11; /*force, mem is only cleared at boot, not restart*/
//    uartByteSpeed = 960; //for vs1063
    uartByteSpeed = 115200/10; //for vs1063
    InitHardware(); /*no longer clears memory in vs1053/63 so preserves const*/
    audioPtr.curFctl = audioPtr.newFctl = 0x15555; //correct for 8kHz 20111117
#if 0
    USEX(FREQCTLH) = (USEX(FREQCTLH) & ~(15|FCH_PLL_LOCKED))
      | (u_int16)(audioPtr.curFctl>>16);
    USEX(FREQCTLL) = audioPtr.curFctl;
    AudioSet(1, 48000U);
    USEX(FREQCTLH) = (USEX(FREQCTLH) & ~(15|FCH_PLL_LOCKED))
      | (u_int16)(audioPtr.newFctl>>16);
    USEX(FREQCTLL) = audioPtr.newFctl;
    audioPtr.curFctl = audioPtr.newFctl;
    audioPtr.changeTimer = -1;
#endif

#if 1
    //2 = drivers, 1 = anapwr
    //USEX(SCI_STATUS) &= ~((1<<SCIST_APDOWN2)|(1<<SCIST_APDOWN1)); // bad
    //USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN2); //drivers first - bad
    USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1); //powers first - best
#endif
    //putstrp("\pstarting");

  #ifndef USE_DEBUG
    /* Enable UART RX interrupt to be able to jump to emulator.
       Allows reflashing without MMC/SD inserted. */
    USEX(INT_ENABLE) = (1<<INT_EN_RX);
  #endif

    /* If not powerup reset, re-enable analog drivers immediately.
       InitHardware() doesn't actually touch powerdowns.
       CheckVolumeAndBass() clears analog powerdown.
     */
    if (!fullReset) {
      //USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN2); //driver powerdown
    }
#ifdef HAS_KEYS
    USEX(TIMER_CONFIG) = clockX;
    *(volatile u_int32 *)(TIMER_T0L) = 6000000/32; /*16Hz->32Hz*/
    USEX(TIMER_ENABLE) |= 1;
    USEX(INT_ENABLE) |= (1<<INT_EN_TIM0);
#endif
    /* Allow DREQ and enable interrupts, VERY IMPORTANT (such as started from SCI)! */
    /* Enable interrupts */
    USEX(INT_GLOB_ENA) =
#ifdef USE_DASH_IN_PLAYER
      (u_int16)AacPatch /*Pull in DASH support for AAC. */
#else
      0
#endif
      ;
    USEX(INT_GLOB_ENA) =
#ifdef HAS_KEYS
      (u_int16)TimerInt /*Pull in timer interrupt - provides a timebase.*/
#else
      0
#endif
      ;
    //USEX(INT_ENABLE) = 0;

    /* wake up MMC! -- If this takes too long, watchdog will bite. */

  #ifdef SAVE_POSITION
    /* Clear the current song number so it will reset between cards. */
    if (USEX(SCI_AICTRL0)) /* 9 words */
	SaveWord(28/*AICTRL0*/, 0); //3 words
  #endif

#if 0
    parametric_x.earSpeakerLevel = 0x3000;
#endif

 cardchanged:
    memsetY(&mmc, 0, sizeof(mmc));
    InitializeMmc(50);
    if (mmc.state != mmcOk || mmc.errors) {
      putword(('!'<<8)|'C'); /*temporary debugs*/
      goto cardchanged;
    }

    // Init File System (FAT)
    /* Detects cards without partition table */
    memset(stream_buffer, 0, 512/2);
    if (InitFileSystem() != 0) {
      putword(('!'<<8)|'F');
      goto cardchanged;
    }
  #ifdef HAS_KEYS
    ui.Key = 0;
    ui.KeyTime = -32767; /*Ignore record button release*/
  #endif
    
#ifdef CTRL3_NO_NUMFILES
	if ((USEX(SCI_AICTRL3) & CTRL3_NO_NUMFILES)) {
	  /* Startup is faster if we don't need to find number of files. */
	  USEX(SCI_AICTRL1) = 0x7fff;
	} else
#endif
	{
	  /* determine number of playable files */
	  USEX(SCI_AICTRL1) = OpenFile(0xffffU);
	  //putword(USEX(SCI_AICTRL1));
	}
#ifdef USE_PRINTABLE_OUTPUT
    /* In decimal format */
    putstrp("\pfat ");
    putdec(fatInfo.totSize);
    putstrp("\p\nfiles ");
    putdec(USEX(SCI_AICTRL1));
    putch('\n');
#else
    #if 1
    /* In hex format */
    putstrp("\pfat ");
    puthex(fatInfo.totSize>>16);
    puthex(fatInfo.totSize);
    putstrp("\p\nfiles ");
    puthex(USEX(SCI_AICTRL1));
    putch('\n');
    #else
    putword(USEX(SCI_AICTRL1));
    putch('\n');
    #endif
#endif


    USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX
#ifdef HAS_KEYS
      |(1<<INT_EN_TIM0)
#endif
      ;
 fileloop:
    /*if (fileMode) */
#ifdef UART_BUFFERED
    if (fileMode) {
	static u_int16 buffer[80];
	register u_int16 *bf = buffer;

	goto prompt;
	while (1) {
	    if (UartFill()) {
		int cmd = UartGetByte();
		*bf = 0;
		if (cmd == '\n' || bf == &buffer[79]) {
		    /* Build a 8.3-character filename */
		    fileName[0] = (buffer[1]<<8) | buffer[2];
		    fileName[1] = (buffer[3]<<8) | buffer[4];
		    fileName[2] = (buffer[5]<<8) | buffer[6];
		    fileName[3] = (buffer[7]<<8) | buffer[8];
		    fileName[4] = (buffer[9]<<8) | buffer[10];
		    fileName[5] = (buffer[11]<<8) | 0x20; /* Archive flag */

		    /* Whenever a command is entered,
		       poll MMC/SD present by giving it a command.
		       But only when idle, i.e. mmcOk. */
		    if (mmc.state == mmcOk && mmc.errors == 0 &&
			MmcCommand(MMC_SET_BLOCKLEN|0x40, 512) != 0) {
			mmc.errors++;
#if defined(USE_PRINTABLE_OUTPUT) || defined(UART_BUFFERED)
			putstrp("\pcard removed\n");
#endif
			goto cardchanged;
		    }
		    switch (buffer[0]) {
		    case 'c':
			/* continuous play mode */
			fileMode = 0;
			goto playloop;
#if 1
		    case 'L':
			/* List files - calls IterateFilesCallback() */
			Disable();
			USEX(SCI_AICTRL3) |= CTRL3_ITERATE;
			OpenFile(0x7fff);
			//USEX(SCI_AICTRL3) &= ~CTRL3_ITERATE; /*turn off! turned off by OpenFile() */
			Enable();
			break;
#endif
#ifdef DELETE_FILE
		    case 'D':  /* 'D'elete by name */
			/* already put the name into the fileName variable */
			fatInfo.dirLine = -1;
			USEX(SCI_AICTRL3) |= CTRL3_BY_NAME;
			OpenFile(0x7fff); /*OpenFile updates AICTRL0*/
			//USEX(SCI_AICTRL3) &= ~CTRL3_BY_NAME; //Cleared by OpenFile()
			if (fatInfo.dirLine >= 0) {
#ifdef USE_PRINTABLE_OUTPUT
			    putstrp("\pDeleting\n");
#endif
			    /* FsDeleteFile() requires dirSector, dirLine,
			       and fragments[] initialized. */
			    FsDeleteFile();
			} else {
#ifdef USE_PRINTABLE_OUTPUT
			    putstrp("\pnot found\n");
#else
			    putstrp("\pnof\n");
#endif
			}
			break;
#endif

		    case 'P':  /* 'P'lay by name */
			/* already put the name into the fileName variable */
			USEX(SCI_AICTRL0) = USEX(SCI_AICTRL1);
			USEX(SCI_AICTRL3) |= CTRL3_BY_NAME;
			OpenFile(0x7fff); /*OpenFile updates AICTRL0 if it finds file*/
			//USEX(SCI_AICTRL3) &= ~CTRL3_BY_NAME; /*cleared by OpenFile() */
			goto checkvalidfile;
		    case 'p': /* 'p'lay by number */
			USEX(SCI_AICTRL0) = atou((void *)&buffer[1]);
		    checkvalidfile:
			if (USEX(SCI_AICTRL0) >= USEX(SCI_AICTRL1)) {
#ifdef USE_PRINTABLE_OUTPUT
			    putstrp("\pnot found ");
			    puthex(USEX(SCI_AICTRL0));
			    putch(' ');
			    putstrpy(fileName);
			    putch('\n');
#else
			    putstrp("\pnof\n");
#endif
			    break;
			}
			goto playloop;
		    }
		    bf = buffer;
		prompt:
#ifdef USE_PRINTABLE_OUTPUT
		    putword(('>'<<8)|' ');
#else
		    ;
#endif
		} else {
		    *bf++ = cmd;
		}
	    }
	    /* If OpenFile found read errors. */
	    if (mmc.state != mmcOk || mmc.errors) {
		goto cardchanged;
	    }
	}
    }

 playloop:
#endif /*UART_BUFFERED*/
    
	if ((USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY)) {
        #ifdef SHUFFLE_PLAY /*shuffle 23 words*/
	  USEX(SCI_AICTRL0) = Shuffle(USEX(SCI_AICTRL1), USEX(SCI_AICTRL0));
        #else
          /* random play (not shuffle!) */
	  USEX(SCI_AICTRL0) = (myrand() % USEX(SCI_AICTRL1));
        #endif
        }
 reopen:        
	/* Wrap from last file to first file */
	if (USEX(SCI_AICTRL0) >= USEX(SCI_AICTRL1)) {
	  USEX(SCI_AICTRL0) -= USEX(SCI_AICTRL1);
	}
	/* Open the file to play. The file number is in AICTRL0. */

    if (OpenFile(USEX(SCI_AICTRL0) &= 0x7fffU) < 0) { //max 32768 files
      PlayCurrentFile();
      
    } else {
		      //putstrp("\p!f");
      USEX(SCI_AICTRL0) = 0; /* file not found, jump to first file */
      goto cardchanged;
    }
    #ifdef SAVE_POSITION
      /* Save current song number to EEPROM */
      SaveWord(28/*AICTRL0*/, USEX(SCI_AICTRL0));
    #endif

      goto fileloop;
  
}


#if 0
void printfile() { /* Must have been opened with OpenFile() */
    spix.size = (fatInfo.fileSize+1)/2; /* convert to words */
    spix.fragment = &fragments[0];
    while (1) {
	spix.sector   = spix.fragment->start /*& 0x7fffffffUL*/;
	spix.fragSize = spix.fragment->size;
	do {
	    s_int16 sz = 512;
	    int i;
	    register u_int16 *p;
	    if (sz > spix.size) {
		sz = spix.size;
	    }
	    ReadDiskSectorTo(spix.sector++, stream_buffer);
	    p = stream_buffer;
	    for (i=0;i<sz;i+=2) {
		putword(*p++);
	    }
	    if (sz & 1) {
		putch(*p>>8);
	    }
	    spix.sz -= sz;
	} while (spix.size != 0 && --spix.fragSize > 0);
	if (spix.size == 0 || (s_int32)spix.fragment->start < 0)
	    break;
	spix.fragment++;
    }
}
#endif
